/*
sudoku - Create and play sudoku

Copyright (C) 2007 Sven Heinzel <Sven.Heinzel@gmx.net>
 
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#pragma once

#ifndef GRIDVIEW_H
#define GRIDVIEW_H

#include "sudoku_gui.h"
#include "DisplayOptions.h"

#include "sudoku/GenerateOptions.h"
#include "sudoku/Grid.h"
#include "sudoku/Solver.h"
#include "sudoku/Generator.h"

class GridView : public QAbstractScrollArea, public Sudoku::Grid
{
    Q_OBJECT

public:
    explicit GridView( QWidget  *pParent = 0 );
    ~GridView();

    inline DisplayOptions GetDisplayOptions() const
    {
        return mDisplayOptions;
    }

    inline double GetZoom() const
    {
        return mZoom;
    }

    inline bool IsEditMode() const
    {
        return mEditMode;
    }

    bool Save( QDomElement  &rParent, bool const  gameMode ) const;
    bool Load( QDomElement const  &rParent );
    void Restart();
    bool Generate( Sudoku::GenerateOptions const  &rOptions );
    void SetDisplayOptions( DisplayOptions const  &rOptions );
    bool SolveOne();
    bool SolveAll();
    void Undo();
    void Redo();
    bool CanUndo() const;
    bool CanRedo() const;
    void DrawSudoku( QPainter  &rPainter, double  width, DisplayOptions const  &rOptions = DisplayOptions() );

Q_SIGNALS:
    void Modified();
    void Message( QString const  &rMessage );
    void ZoomChanged( double const  zoom );
    void EditModeChanged( bool const  enabled );

public Q_SLOTS:
    inline void SetZoom( double const  zoom )
    {
        double const  new_zoom = qMax( 1.0, qMin( zoom, 7.0 ) );
        if( mZoom != new_zoom ) {
            mZoom = new_zoom;
            UpdateScrollBars();
            emit ZoomChanged( mZoom );
            viewport() -> update();
        }
    }

    inline void SetEditMode( bool const  enabled )
    {
        if( mEditMode != enabled ) {
            mEditMode = enabled;
            emit EditModeChanged( mEditMode );
            viewport() -> update();
        }
    }

protected:
    void paintEvent( QPaintEvent  *pEvent );
    void resizeEvent( QResizeEvent  *pEvent );
    void mousePressEvent( QMouseEvent  *pEvent );
    void mouseMoveEvent( QMouseEvent  *pEvent );
    void keyPressEvent( QKeyEvent  *pEvent );
    void wheelEvent( QWheelEvent  *pEvent );

    inline void Changed()
    {
        emit Modified();
        viewport() -> update();
    }

private:
    inline double CalcDisplayWidth( int  dpiX, int  dpiY ) const
    {
        double const  dpi_x = static_cast< double >( dpiX );
        double const  dpi_y = static_cast< double >( dpiY );
        double const  size_width  = GetWidth() / dpi_x * 25.4;
        double const  size_height = GetHeight() / dpi_y * 25.4;

        // Use the minimum size to fit the complete area in the widget
        return qMin( size_width, size_height );
    }

    inline double GetSize() const
    {
        double const  w = GetWidth();
        double const  h = GetHeight();
        return qMin( w, h );
    }

    inline double GetXOffset() const
    {
        double const  offset = ( viewport() -> width() - 2 * mBorder * mZoom - GetSize() ) / 2.0;
        return mBorder * mZoom + qMax( 0.0, offset );
    }

    inline double GetYOffset() const
    {
        double const  offset = ( viewport() -> height() - 2 * mBorder * mZoom - GetSize() ) / 2.0;
        return mBorder * mZoom + qMax( 0.0, offset );
    }

    inline double GetX() const
    {
        return GetXOffset() - horizontalScrollBar() -> value();
    }

    inline double GetY() const
    {
        return GetYOffset() - verticalScrollBar() -> value();
    }

    inline double GetWidth() const
    {
        return ( viewport() -> width() - 2 * mBorder ) * mZoom;
    }

    inline double GetHeight() const
    {
        return ( viewport() -> height() - 2 * mBorder ) * mZoom;
    }

    inline double GetXPos( int const  x ) const
    {
        double const  size = GetSize() / static_cast< double >( GetCount() );
        return ( static_cast< double >( x ) - GetX() ) / size;
    }

    inline double GetYPos( int const  y ) const
    {
        double const  size = GetSize() / static_cast< double >( GetCount() );
        return ( static_cast< double >( y ) - GetY() ) / size;
    }

    void UpdateLines();
    void UpdateScrollBars();
    int KeyToNumber( QKeyEvent  *pEvent ) const;
    void SetCurrent( int  x, int  y );
    void PushUndo();
    void ResetUndo();

    static double const  mUnit;
    double          mZoom;
    int             mBorder;
    QPoint          mCurrent;
    bool            mEditMode;
    QPicture        mSudokuLines;
    DisplayOptions  mDisplayOptions;
    QStringList     mSymbolList;
    Sudoku::SolveOptions    mSolveOptions;
    QStack< Sudoku::Grid >  mUndoStack;
    QStack< Sudoku::Grid >  mRedoStack;
};

#endif
