/*
sudoku - Create and play sudoku

Copyright (C) 2007 Sven Heinzel <Sven.Heinzel@gmx.net>
 
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "sudoku_gui.h"
#include "DockWidgets.h"

GenerateDockWidget::GenerateDockWidget(
    GridView         *pGridView,
    QWidget          *pParent,
    Qt::WindowFlags  flags
)
    : BaseDockWidget( pParent, flags )
    , mpGenerateWidget( 0 )
{
    mpGenerateWidget = new GenerateWidget( Sudoku::GenerateOptions(), this );
    QPushButton  *p_button = new QPushButton( tr( "Generate" ), this );
    connect( p_button, SIGNAL( clicked() ), this, SLOT( Generate() ) );

    QVBoxLayout  *p_layout = new QVBoxLayout( this );
    p_layout -> addWidget( mpGenerateWidget );
    p_layout -> addWidget( p_button );
    p_layout -> addStretch( 1 );

    // Initial setup
    SetGridView( pGridView );
}

void GenerateDockWidget::Generate()
{
    // Generate sudoku
    if( mpGridView ) mpGridView -> Generate( mpGenerateWidget -> GetGenerateOptions() );
}

SolveDockWidget::SolveDockWidget(
    GridView         *pGridView,
    QWidget          *pParent,
    Qt::WindowFlags  flags
)
    : BaseDockWidget( pParent, flags )
    , mpCheckShowAllowed( 0 )
    , mpCheckShowSimplifiedAllowed( 0 )
    , mpCheckShowTips( 0 )
{
    QGroupBox  *p_group = new QGroupBox( tr( "Help" ), this );
    mpCheckShowAllowed = new QCheckBox( tr( "Show allowed numbers" ), p_group );
    mpCheckShowSimplifiedAllowed = new QCheckBox( tr( "Simplify allowed numbers" ), p_group );
    mpCheckShowTips = new QCheckBox( tr( "Show solution tips" ), p_group );
    connect( mpCheckShowAllowed, SIGNAL( toggled( bool ) ), this, SLOT( ChangeDisplay() ) );
    connect( mpCheckShowSimplifiedAllowed, SIGNAL( toggled( bool ) ), this, SLOT( ChangeDisplay() ) );
    connect( mpCheckShowTips, SIGNAL( toggled( bool ) ), this, SLOT( ChangeDisplay() ) );

    QVBoxLayout  *p_group_layout = new QVBoxLayout( p_group );
    p_group_layout -> addWidget( mpCheckShowAllowed );
    p_group_layout -> addWidget( mpCheckShowSimplifiedAllowed );
    p_group_layout -> addWidget( mpCheckShowTips );

    QPushButton  *p_button_one = new QPushButton( tr( "Solve one" ), this );
    QPushButton  *p_button_all = new QPushButton( tr( "Solve all" ), this );
    connect( p_button_one, SIGNAL( clicked() ), this, SLOT( SolveOne() ) );
    connect( p_button_all, SIGNAL( clicked() ), this, SLOT( SolveAll() ) );

    QVBoxLayout  *p_layout = new QVBoxLayout( this );
    p_layout -> addWidget( p_group );
    p_layout -> addWidget( p_button_one );
    p_layout -> addWidget( p_button_all );
    p_layout -> addStretch( 1 );

    // Initial setup
    SetGridView( pGridView );
}

void SolveDockWidget::ChangeDisplay()
{
    // Change display options
    if( mpGridView ) {
        DisplayOptions  options = mpGridView -> GetDisplayOptions();
        options . SetVisible( DisplayOptions::ShowAllowed, mpCheckShowAllowed -> isChecked() );
        options . SetVisible( DisplayOptions::ShowSimplifiedAllowed, mpCheckShowSimplifiedAllowed -> isChecked() );
        options . SetVisible( DisplayOptions::ShowTips, mpCheckShowTips -> isChecked() );
        mpGridView -> SetDisplayOptions( options );
    }
}

void SolveDockWidget::UpdateDisplay()
{
    // Update widgets
    if( mpGridView ) {
        DisplayOptions const  options = mpGridView -> GetDisplayOptions();
        mpCheckShowAllowed -> setChecked( options . IsVisible( DisplayOptions::ShowAllowed ) );
        mpCheckShowSimplifiedAllowed -> setChecked( options . IsVisible( DisplayOptions::ShowSimplifiedAllowed ) );
        mpCheckShowSimplifiedAllowed -> setEnabled( options . IsVisible( DisplayOptions::ShowAllowed ) );
        mpCheckShowTips -> setChecked( options . IsVisible( DisplayOptions::ShowTips ) );
    }
}

void SolveDockWidget::SolveOne()
{
    // Solve one
    if( mpGridView ) mpGridView -> SolveOne();
}

void SolveDockWidget::SolveAll()
{
    // Solve all
    if( mpGridView ) mpGridView -> SolveAll();
}

AnalyseDockWidget::AnalyseDockWidget(
    GridView         *pGridView,
    QWidget          *pParent,
    Qt::WindowFlags  flags
)
    : BaseDockWidget( pParent, flags )
    , mpLabelSolution( 0 )
    , mpLabelPairs( 0 )
    , mpLabelBacktrack( 0 )
{
    QGroupBox  *p_group_solution = new QGroupBox( tr( "Solvable" ), this );
    QLabel  *p_label_solution = new QLabel( tr( "Solution:" ), this );
    mpLabelSolution = new QLabel( this );
    mpLabelSolution -> setAutoFillBackground( true );

    QGridLayout  *p_group_solution_layout = new QGridLayout( p_group_solution );
    p_group_solution_layout -> addWidget( p_label_solution, 0, 0 );
    p_group_solution_layout -> addWidget( mpLabelSolution, 0, 1 );

    QGroupBox  *p_group_options = new QGroupBox( tr( "Solve options" ), this );
    QLabel  *p_label_pairs = new QLabel( tr( "Maximum pair:" ), this );
    mpLabelPairs = new QLabel( this );
    QLabel  *p_label_backtrack = new QLabel( tr( "Backtrack:" ), this );
    mpLabelBacktrack = new QLabel( this );

    QGridLayout  *p_group_options_layout = new QGridLayout( p_group_options );
    p_group_options_layout -> addWidget( p_label_pairs, 0, 0 );
    p_group_options_layout -> addWidget( mpLabelPairs, 0, 1 );
    p_group_options_layout -> addWidget( p_label_backtrack, 1, 0 );
    p_group_options_layout -> addWidget( mpLabelBacktrack, 1, 1 );

    QVBoxLayout  *p_layout = new QVBoxLayout( this );
    p_layout -> addWidget( p_group_solution );
    p_layout -> addWidget( p_group_options );
    p_layout -> addStretch( 1 );

    // Initial setup
    SetGridView( pGridView );
}

void AnalyseDockWidget::UpdateDisplay()
{
    // Update widgets
    if( mpGridView ) {
        Sudoku::Solver  solver;
        Sudoku::SolveResult const  result = solver . Analyse( *mpGridView );
        QString  solution;
        switch( result . GetSolution() ) {
            default:
            case Sudoku::SolveResult::NoSolution:
                solution = tr( "None" );
                break;
            case Sudoku::SolveResult::UniqueSolution:
                solution = tr( "Unique" );
                break;
            case Sudoku::SolveResult::MultipleSolutions:
                solution = tr( "Multiple" );
                break;
        }
        mpLabelSolution -> setText( solution );
        mpLabelPairs -> setText( result . GetMaxPairSolving() > 1 ? QString::number( result . GetMaxPairSolving() ) : tr( "-" ) );
        mpLabelBacktrack -> setText( result . IsBacktrackAllowed() ? tr( "Used" ) : tr( "-" ) );
    }
}
