/*
sudoku - Create and play sudoku

Copyright (C) 2007 Sven Heinzel <Sven.Heinzel@gmx.net>
 
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "sudoku_gui.h"
#include "GridView.h"

#include "sudoku/System.h"
#include "sudoku/Grid.h"
#include "sudoku/Solver.h"
#include "sudoku/Version.h"

double const  GridView::mUnit = 100.0;

GridView::GridView(
    QWidget  *pParent
)
    : QAbstractScrollArea( pParent )
    , mZoom( 1.0 )
    , mBorder( 10 )
    , mCurrent( 0, 0 )
    , mEditMode( false )
    , mDisplayOptions( DisplayOptions::ShowFields | DisplayOptions::ShowActive )
    , mSolveOptions( Sudoku::SolveOptions::all() )
{
    // Set minimum size to fit all fields of a standard sized sudoku
    setMinimumSize( 2 * mBorder + 9 * 30, 2 * mBorder + 9 * 30 );
    setFocusPolicy( Qt::StrongFocus );

    mSymbolList << "0" << "1" << "2" << "3" << "4" << "5" << "6" << "7" << "8" << "9"
                << "A" << "B" << "C" << "D" << "E" << "F" << "G" << "H" << "I" << "J"
                << "K" << "L" << "M" << "N" << "O" << "P" << "Q" << "R" << "S" << "T"
                << "U" << "V" << "W" << "X" << "Y" << "Z" << "@" << "#" << "$" << "%"
                << "*" << "=" << "a" << "b" << "c" << "d" << "e" << "f" << "g" << "h"
                << "j" << "k" << "m" << "n" << "o" << "p" << "q" << "r" << "s" << "t"
                << "u" << "v" << "w" << "y" << "z";

    UpdateLines();
    UpdateScrollBars();
    ResetUndo();
}

GridView::~GridView()
{
}

void GridView::UpdateLines()
{
    QPen  pen;
    pen . setWidthF( 1.0 );
    pen . setJoinStyle( Qt::MiterJoin );
    pen . setCapStyle( Qt::FlatCap );
    QPen  bold_pen;
    bold_pen . setWidthF( 5.0 );
    bold_pen . setJoinStyle( Qt::MiterJoin );
    bold_pen . setCapStyle( Qt::FlatCap );
    int const  c = GetCount();

    QPainter  painter;
    if( painter . begin( &mSudokuLines ) ) {
        // Small lines
        painter . setPen( pen );
        for( int  i = 1; i < c; ++i ) {
            painter . drawLine( QLineF( 0.0, i * mUnit, c * mUnit, i * mUnit ) );
            painter . drawLine( QLineF( i * mUnit, 0.0, i * mUnit, c * mUnit ) );
        }

        // Border
        painter . setPen( bold_pen );
        painter . drawRect( QRectF( 0.0, 0.0, c * mUnit, c * mUnit ) );

        // Big lines
        for( int  y = 0; y < c; ++y ) {
            for( int  x = 0; x < c; ++x ) {
                if( ( ( x + 1 ) < c ) && BoxIndex( x, y ) != BoxIndex( x + 1, y ) ) {
                    painter . drawLine( QLineF( ( x + 1 ) * mUnit, y * mUnit, ( x + 1 ) * mUnit, ( y + 1 ) * mUnit ) );
                }
                if( ( ( y + 1 ) < c ) && BoxIndex( x, y ) != BoxIndex( x, y + 1 ) ) {
                    painter . drawLine( QLineF( x * mUnit, ( y + 1 ) * mUnit, ( x + 1 ) * mUnit, ( y + 1 ) * mUnit ) );
                }
            }
        }

        painter . end();
    }
}

void GridView::UpdateScrollBars()
{
    setVerticalScrollBarPolicy( mZoom > 1.0 ? Qt::ScrollBarAlwaysOn : Qt::ScrollBarAlwaysOff );
    setHorizontalScrollBarPolicy( mZoom > 1.0 ? Qt::ScrollBarAlwaysOn : Qt::ScrollBarAlwaysOff );
    int const  single_step = 1;
    int const  page_step = static_cast< int >( GetSize() / GetCount() + 0.5 );
    horizontalScrollBar() -> setRange( 0, static_cast< int >( GetSize() - viewport() -> width() + 2 * mBorder * mZoom + 0.5 ) );
    horizontalScrollBar() -> setSingleStep( single_step );
    horizontalScrollBar() -> setPageStep( page_step );
    verticalScrollBar() -> setRange( 0, static_cast< int >( GetSize() - viewport() -> height() + 2 * mBorder * mZoom + 0.5 ) );
    verticalScrollBar() -> setSingleStep( single_step );
    verticalScrollBar() -> setPageStep( page_step );
}

void GridView::SetCurrent(
    int  x,
    int  y
)
{
    // Check if new position is valid
    if( ( x >= 0 ) && ( x < GetCount() ) && ( y >= 0 ) && ( y < GetCount() ) ) {
        mCurrent = QPoint( x, y );
        viewport() -> update();
    }
}

void GridView::mousePressEvent(
    QMouseEvent  *pEvent
)
{
    // Just call the move event and handle it there
    mouseMoveEvent( pEvent );
}

void GridView::mouseMoveEvent(
    QMouseEvent  *pEvent
)
{
    if( pEvent ) {
        if( pEvent -> buttons() & Qt::LeftButton ) {
            // On left button press/move set new position
            SetCurrent( static_cast< int >( GetXPos( pEvent -> x() ) ), static_cast< int >( GetYPos( pEvent -> y() ) ) );
        }
    }
}

int GridView::KeyToNumber(
    QKeyEvent  *pEvent
) const
{
    if( pEvent ) {
        QString const  key_text = pEvent -> text();
        for( int  i = 0; i < mSymbolList . size(); ++i ) {
            if( mSymbolList[ i ] == key_text ) return i;
        }
    }

    // Nothing found
    return -1;
}

void GridView::keyPressEvent(
    QKeyEvent  *pEvent
)
{
    if( pEvent ) {
        int   x = mCurrent . x();
        int   y = mCurrent . y();
        bool const  fixed = IsFixed( x, y );
        int const   key = pEvent -> key();
        int const   number = KeyToNumber( pEvent );

        // Handle keys

        if( key == Qt::Key_Left ) {
            --x;
        } else if( key == Qt::Key_Right ) {
            ++x;
        } else if( key == Qt::Key_Up ) {
            --y;
        } else if( key == Qt::Key_Down ) {
            ++y;
        } else if( ( number >= 0 ) && ( number <= GetCount() ) ) {
            if( IsNumberAllowed( x, y, number ) && ( GetNumber( x, y ) != number ) ) {
                PushUndo();

                // Set/clear field
                if( mEditMode ) SetFixed( x, y, false );
                SetNumber( x, y, number );
                SetFixed( x, y, fixed );
                Changed();
            }
        } else if( ( key == Qt::Key_Delete ) || ( key == Qt::Key_Backspace ) ) {
            if( GetNumber( x, y ) != 0 ) {
                PushUndo();

                // Clear field
                if( mEditMode ) SetFixed( x, y, false );
                SetNumber( x, y, 0 );
                SetFixed( x, y, fixed );
                Changed();
            }
        } else if( key == Qt::Key_Space ) {
            if( mEditMode ) {
                PushUndo();

                // Toggle fixed
                SetFixed( x, y, ! fixed );
                Changed();
            }
        } else {
            // If key is unhandled let the base class handle it
            pEvent -> ignore();
        }

        // Check for overflow at the borders
        int const  c = GetCount() - 1;
        if( x < 0 ) {
            x = c;
            --y;
        }
        if( x > c ) {
            x = 0;
            ++y;
        }
        if( y < 0 ) y = c;
        if( y > c ) y = 0;

        // Set new position (this also causes an update)
        SetCurrent( x, y );
    }
}

void GridView::paintEvent(
    QPaintEvent  *pEvent
)
{
    QPainter  painter;
    if( painter . begin( viewport() ) ) {
        // Fill background
        painter . fillRect( contentsRect(), palette() . dark() );

        // Set clipping
        if( pEvent ) painter . setClipRect( pEvent -> rect() );

        // Translate the painter
        painter . translate( GetX(), GetY() );

        // Draw the sudoku
        DrawSudoku( painter, CalcDisplayWidth( painter . device() -> logicalDpiX(), painter . device() -> logicalDpiY() ), mDisplayOptions );

        painter . end();
    }
}

void GridView::resizeEvent(
    QResizeEvent  *pEvent
)
{
    if( pEvent ) UpdateScrollBars();
}

void GridView::wheelEvent(
    QWheelEvent  *pEvent
)
{
    if( pEvent ) {
        double const  x = GetXPos( pEvent -> x() );
        double const  y = GetYPos( pEvent -> y() );
        double const  factor = 1.1;
        SetZoom( pEvent -> delta() > 0 ? mZoom * factor : mZoom / factor );
        double const  scale = GetSize() / static_cast< double >( GetCount() );
        horizontalScrollBar() -> setValue( static_cast< int >( GetXOffset() + x * scale - pEvent -> x() + 0.5 ) );
        verticalScrollBar() -> setValue( static_cast< int >( GetYOffset() + y * scale - pEvent -> y() + 0.5 ) );
    }
}

void GridView::DrawSudoku(
    QPainter              &rPainter,
    double                width,
    DisplayOptions const  &rOptions
)
{
    // Set scaling
    double const  size = GetCount() * mUnit;
    double const  scale_x = static_cast< double >( rPainter . device() -> logicalDpiX() ) / ( 25.4 * size / width );
    double const  scale_y = static_cast< double >( rPainter . device() -> logicalDpiY() ) / ( 25.4 * size / width );
    rPainter . scale( scale_x, scale_y );

    // Background
    rPainter . fillRect( QRectF( 0.0, 0.0, size, size ), QColor( 255, 255, 255 ) );

    // Configure painter options
    rPainter . setRenderHint( QPainter::Antialiasing );
    rPainter . setRenderHint( QPainter::TextAntialiasing );
    rPainter . setRenderHint( QPainter::SmoothPixmapTransform );

    // Prepare brushes etc.
    QBrush const  fg = QApplication::palette() . windowText();
    QBrush const  dis_bg = QColor( 240, 230, 220 );
    QBrush const  high_bg = QApplication::palette() . highlight();
    QBrush const  high_fg = QApplication::palette() . highlightedText();
    QBrush const  tip_bg = QColor( 130, 230, 130 );
    QBrush const  tip_fg = QColor( 250, 150, 150 );
    int const     text_option = Qt::AlignCenter;

    // Analyse only if tips should be shown
    Sudoku::Grid  analyse_grid( *this );
    if( rOptions . IsVisible( DisplayOptions::ShowTips ) ) {
        analyse_grid = Sudoku::Solver() . FindTips( *this, rOptions . IsVisible( DisplayOptions::ShowSimplifiedAllowed ) );
    }

    QFont  font_big = QFont( "Times" );
    //font_big . setPointSizeF( 70.0 );
    font_big . setPixelSize( 80 );
    font_big . setStyleStrategy( QFont::PreferQuality );
    QFont  font_small = QFont( "Times" );
    //font_small . setPointSizeF( 20.0 );
    font_small . setPixelSize( 20 );
    font_small . setStyleStrategy( QFont::PreferQuality );

    Sudoku::Allowed const  allowed = GetAllowed();
    Sudoku::Allowed const  allowed_simplified = rOptions . IsVisible( DisplayOptions::ShowSimplifiedAllowed )
        ? Sudoku::Solver() . SimplifyAllowed( *this )
        : allowed;

    int const  current_x = mCurrent . x();
    int const  current_y = mCurrent . y();
    int const  small_q = static_cast< int >( ceil( sqrt( static_cast< double >( GetCount() ) ) - 0.01 ) );
    double const  small_size = mUnit / small_q;
    bool const    show_small = small_size * scale_x > 12.0;
    bool          show = false;
    //QRectF  view_rect( -GetX() / scale_x, -GetY() / scale_y, viewport() -> width() / scale_x, viewport() -> height() / scale_y );
#if QT_VERSION >= 0x040800
    // Available for Qt 4.8.0 and later
    QRectF  view_rect = rPainter . clipBoundingRect();
#else
    QRectF  view_rect = rPainter . clipRegion() . boundingRect();
#endif
    for( int  y = 0; y < GetCount(); ++y ) {
        for( int  x = 0; x < GetCount(); ++x ) {
            // Check if field is in view
            QRectF const  rect = QRectF( x * mUnit, y * mUnit, mUnit, mUnit );
            if( view_rect . isValid() && !view_rect . intersects( rect ) ) continue;

            int const     b = BoxIndex( x, y );
            int const     num = GetNumber( x, y );

            if( IsFixed( x, y ) ) {
                // Fixed
                show = rOptions . IsVisible( DisplayOptions::ShowFixed );
                if( show ) {
                    rPainter . fillRect( rect, dis_bg );
                    rPainter . setPen( fg . color() );
                }
            } else {
                // Normal
                show = rOptions . IsVisible( DisplayOptions::ShowNonFixed );
                if( show ) {
                    rPainter . setPen( fg . color() );
                }
            }

            if( rOptions . IsVisible( DisplayOptions::ShowTips ) && ( num == 0 ) && ( analyse_grid . GetNumber( x, y ) != 0 ) ) {
                // Tips
                rPainter . fillRect( rect, tip_bg );
            }

            if( rOptions . IsVisible( DisplayOptions::ShowActive ) && hasFocus() && ( current_y == y ) && ( current_x == x ) ) {
                // Focus
                rPainter . fillRect( rect, high_bg );
                rPainter . setPen( high_fg . color() );
            }

            if( num == 0 ) {
                if( rOptions . IsVisible( DisplayOptions::ShowAllowed ) ) {
                    // Small numbers in help mode for free fields
                    rPainter . setFont( font_small );
                    QPen const  text_pen = rPainter . pen();
                    QPen const  tip_pen = QPen( tip_fg . color() );
                    for( int  n = 1; n <= GetCount(); ++n ) {
                        if( allowed . IsNumberAllowed( x, y, b, n ) ) {
                            int const  sx = ( n - 1 ) % small_q;
                            int const  sy = ( n - 1 ) / small_q;
                            if( show_small ) {
                                // Display text
                                rPainter . setPen( allowed_simplified . IsNumberAllowed( x, y, b, n ) ? text_pen : tip_pen );
                                rPainter . drawText( QRectF( x * mUnit + sx * small_size, y * mUnit + sy * small_size, small_size, small_size ), text_option, QString::number( n ) );
                            } else {
                                // Just show small boxes
                                QColor const  fill_col = allowed_simplified . IsNumberAllowed( x, y, b, n ) ? text_pen . color() : tip_pen . color();
                                rPainter . fillRect( QRectF( x * mUnit + ( sx + 0.4 ) * small_size, y * mUnit + ( sy + 0.4 ) * small_size, small_size * 0.2, small_size * 0.2 ), fill_col );
                            }
                        }
                    }
                    rPainter . setPen( text_pen );
                }
            } else {
                // Big number
                if( show ) {
                    rPainter . setFont( font_big );
                    rPainter . drawText( rect, text_option, QString::number( num ) );
                }
            }
        }
    }

    // Border
    double const  screen_scale_x = static_cast< double >( mSudokuLines . logicalDpiX() ) / static_cast< double >( rPainter . device() -> logicalDpiX() );
    double const  screen_scale_y = static_cast< double >( mSudokuLines . logicalDpiY() ) / static_cast< double >( rPainter . device() -> logicalDpiY() );
    rPainter . scale( screen_scale_x, screen_scale_y );
    rPainter . drawPicture( QPointF( 0, 0 ), mSudokuLines );
}

bool GridView::Save(
    QDomElement  &rParent,
    bool const   gameMode
) const
{
    QDomDocument  doc = rParent . ownerDocument();

    if( doc . isNull() ) {
        // A document is needed to create sub tags
        return false;
    }

    // Layout
    QDomElement  layout = doc . createElement( "layout" );
    for( int  y = 0; y < GetCount(); ++y ) {
        for( int  x = 0; x < GetCount(); ++x ) {
            QDomElement  tag = doc . createElement( "box" );
            tag . setAttribute( "x", x );
            tag . setAttribute( "y", y );
            QDomText  val = doc . createTextNode( QString::number( BoxIndex( x, y ) ) );
            tag . appendChild( val );
            layout . appendChild( tag );
        }
    }
    rParent . appendChild( layout );

    // Fill in numbers
    QDomElement  grid = doc . createElement( "grid" );
    for( int  y = 0; y < GetCount(); ++y ) {
        for( int  x = 0; x < GetCount(); ++x ) {
            if( GetNumber( x, y ) ) {
                // In field mode only save fixed fields
                if( gameMode || IsFixed( x, y ) ) {
                    QDomElement  tag = doc . createElement( "field" );
                    tag . setAttribute( "x", x );
                    tag . setAttribute( "y", y );
                    tag . setAttribute( "fixed", IsFixed( x, y ) );
                    QDomText  val = doc . createTextNode( QString::number( GetNumber( x, y ) ) );
                    tag . appendChild( val );
                    grid . appendChild( tag );
                }
            }
        }
    }
    rParent . appendChild( grid );

    return true;
}

bool GridView::Load(
    QDomElement const  &rParent
)
{
    // Clear (but keep current layout)
    Reset();

    QDomElement  field_legacy = rParent . firstChildElement( "field" );
    if( ! field_legacy . isNull() ) {
        // Legacy XML load
        Resize( 9 );

        // Read all fields
        while( ! field_legacy . isNull() ) {
            int   x = field_legacy . attribute( "x", "-1" ) . toInt();
            int   y = field_legacy . attribute( "y", "-1" ) . toInt();
            int   val = field_legacy . text() . toInt();
            bool  fixed = field_legacy . attribute( "fixed", "0" ) . toInt();

            SetNumber( x, y, val );
            SetFixed( x, y, fixed );

            field_legacy = field_legacy . nextSiblingElement( "field" );
        }
    } else {
        // New grid with variable size and layout
        int  count = rParent . attribute( "count", "9" ) . toInt();

        // Parse layout tag
        QDomElement  layout = rParent . firstChildElement( "layout" );
        if( ! layout . isNull() ) {
            QDomElement  box = layout . firstChildElement( "box" );

            // Read all layout box numbers
            while( ! box . isNull() ) {
                int   x = box . attribute( "x", "-1" ) . toInt();
                int   y = box . attribute( "y", "-1" ) . toInt();
                int   b = box . text() . toInt();

                //SetNumber( x, y, val );
                //SetFixed( x, y, fixed );

                box = box . nextSiblingElement( "box" );
            }
        }
        // FIXME Apply layout instead
        Resize( count );

        // Parse grid tag
        QDomElement  grid = rParent . firstChildElement( "grid" );
        if( ! grid . isNull() ) {
            QDomElement  field = grid . firstChildElement( "field" );

            // Read all fields
            while( ! field . isNull() ) {
                int   x = field . attribute( "x", "-1" ) . toInt();
                int   y = field . attribute( "y", "-1" ) . toInt();
                int   val = field . text() . toInt();
                bool  fixed = field . attribute( "fixed", "0" ) . toInt();

                SetNumber( x, y, val );
                SetFixed( x, y, fixed );

                field = field . nextSiblingElement( "field" );
            }
        }
    }

    // Check solutions
    Sudoku::Solver  solver( Sudoku::SolveOptions::all() );
    Sudoku::SolveResult::Solution const  solution = solver . Analyse( *this ) . GetSolution();
    if( solution == Sudoku::SolveResult::NoSolution ) {
        QMessageBox::warning( this, tr( "Sudoku warning" ), tr( "This sudoku has no solution." ), QMessageBox::Ok, 0, 0 );
    } else if( solution == Sudoku::SolveResult::MultipleSolutions ) {
        QMessageBox::warning( this, tr( "Sudoku warning" ), tr( "This sudoku has multiple solutions." ), QMessageBox::Ok, 0, 0 );
    }

    Changed();

    return true;
}

bool GridView::Generate(
    Sudoku::GenerateOptions const  &rOptions
)
{
    PushUndo();

    // Try to generate a new sudoku
    QApplication::setOverrideCursor( QCursor( Qt::WaitCursor ) );
    uint64_t const  start = Sudoku::Timer::Nanoseconds();
    Sudoku::Generator  generator( rOptions );
    bool const  ok = generator . Generate( *this );
    uint64_t const  end = Sudoku::Timer::Nanoseconds();
    double const  dur = static_cast< double >( end - start ) / 1.0e6;
    QApplication::restoreOverrideCursor();

    // If generating failed show a message
    if( ! ok ) QMessageBox::warning( this, tr( "Generate" ), tr( "Could not generate a new sudoku!" ), QMessageBox::Ok, 0, 0 );
    else emit Message( tr( "Generating took %L1 ms." ) . arg( dur, 0, 'f', 3 ) );

    Changed();

    return ok;
}

void GridView::SetDisplayOptions(
    DisplayOptions const  &rOptions
)
{
    if( mDisplayOptions != rOptions ) {
        mDisplayOptions = rOptions;
        Changed();
    }
}

void GridView::Restart()
{
    PushUndo();

    Reset( false );
    Changed();
}

bool GridView::SolveOne()
{
    PushUndo();

    uint64_t const  start = Sudoku::Timer::Nanoseconds();
    Sudoku::Solver  solver( mSolveOptions );
    bool  found = solver . SolveOne( *this );
    uint64_t const  end = Sudoku::Timer::Nanoseconds();
    double const  dur = static_cast< double >( end - start ) / 1.0e6;

    if( ! found ) QMessageBox::warning( this, tr( "Solve" ), tr( "Could not find any number." ), QMessageBox::Ok, 0, 0 );
    else emit Message( tr( "Solving took %L1 ms." ) . arg( dur, 0, 'f', 3 ) );
    Changed();

    return found;
}

bool GridView::SolveAll()
{
    PushUndo();

    uint64_t const  start = Sudoku::Timer::Nanoseconds();
    Sudoku::Solver  solver( mSolveOptions );
    bool  complete = solver . SolveAll( *this );
    uint64_t const  end = Sudoku::Timer::Nanoseconds();
    double const  dur = static_cast< double >( end - start ) / 1.0e6;

    if( ! complete ) QMessageBox::warning( this, tr( "Solve" ), tr( "Could not find any more numbers!" ), QMessageBox::Ok, 0, 0 );
    else emit Message( tr( "Solving took %L1 ms." ) . arg( dur, 0, 'f', 3 ) );
    Changed();

    return complete;
}

void GridView::PushUndo()
{
    // Remove redos
    mRedoStack . clear();

    // Append new undo step
    Sudoku::Grid  *p_ourself = this;
    mUndoStack . push( *p_ourself );
}

void GridView::ResetUndo()
{
    mUndoStack . clear();
    mRedoStack . clear();
}

void GridView::Undo()
{
    if( CanUndo() ) {
        // Push current state on redo stack and restore from undo stack
        Sudoku::Grid  *p_ourself = this;
        mRedoStack . push( *p_ourself );
        *p_ourself = mUndoStack . pop();
        Changed();
    }
}

void GridView::Redo()
{
    if( CanRedo() ) {
        // Push current state on undo stack and restore from redo stack
        Sudoku::Grid  *p_ourself = this;
        mUndoStack . push( *p_ourself );
        *p_ourself = mRedoStack . pop();
        Changed();
    }
}

bool GridView::CanUndo() const
{
    return ! mUndoStack . isEmpty();
}

bool GridView::CanRedo() const
{
    return ! mRedoStack . isEmpty();
}
