/*
sudoku - Create and play sudoku

Copyright (C) 2007 Sven Heinzel <Sven.Heinzel@gmx.net>
 
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "sudoku_gui.h"
#include "MainWindow.h"
#include "DockWidgets.h"
#include "GenerateDialog.h"

#include "sudoku/Version.h"

MainWindow::MainWindow(
    QWidget          *pParent,
    Qt::WindowFlags  flags
)
    : QMainWindow( pParent, flags )
    , mpMenuBar( 0 )
    , mpLabelProgress( 0 )
    , mpGridView( 0 )
    , mpMenuToolbars( 0 )
    , mRecentMapper( this )
    , mStateVersion( -1 )
    , mPrinter( QPrinter::HighResolution )
    , mGenerateOptions()
    , mLastPath( QDir::homePath() )
{
    LoadConfig();
    setWindowTitle( "Sudoku" );

    // Init
    Init();
    InitActions();
    InitToolBars();
    InitDockWidgets();
    InitMenus();
    InitWidgets();

    // Restore old state
    restoreState( mState, mStateVersion );
}

MainWindow::~MainWindow()
{
    SaveConfig();
    delete mpMenuToolbars;
}

void MainWindow::Init()
{
    // Create widgets
    mpMenuBar = new QMenuBar( this );
    mpGridView = new GridView( this );
    mpLabelProgress = new QLabel( this );
}

void MainWindow::InitActions()
{
    // Create actions
    //           key             text                           icon                         checkable shortcut              status tip                                          target slot
    SetupAction( "generate",     tr( "&Generate..." ),              QIcon( ":/data/new.png" ),   false,    QKeySequence::New,    tr( "Generates a new sudoku" ),                     this,  SLOT( SudokuGenerate() ) );
    SetupAction( "restart",      tr( "&Restart" ),                  QIcon(),                     false,    tr( "Ctrl+R" ),       tr( "Restarts the sudoku" ),                        this,  SLOT( SudokuRestart() ) );
    SetupAction( "open",         tr( "&Open..." ),                  QIcon( ":/data/open.png" ),  false,    QKeySequence::Open,   tr( "Opens an existing sudoku" ),                   this,  SLOT( SudokuOpen() ) );
    SetupAction( "save",         tr( "&Save" ),                     QIcon( ":/data/save.png" ),  false,    QKeySequence::Save,   tr( "Saves the sudoku" ),                           this,  SLOT( SudokuSave() ) );
    SetupAction( "saveas",       tr( "Save &as..." ),               QIcon(),                     false,    QKeySequence::SaveAs, tr( "Saves the sudoku under a new name" ),          this,  SLOT( SudokuSaveAs() ) );
    //SetupAction( "export",       tr( "&Export..." ),                QIcon(),                     false,    0,                    tr( "Exports the sudoku" ),                         this,  SLOT( SudokuExport() ) );
    SetupAction( "print",        tr( "&Print..." ),                 QIcon( ":/data/print.png" ), false,    QKeySequence::Print,  tr( "Prints the sudoku" ),                          this,  SLOT( SudokuPrint() ) );
    SetupAction( "printsetup",   tr( "Printer setup..." ),          QIcon(),                     false,    0,                    tr( "Configures the printer" ),                     this,  SLOT( SudokuPrinterSetup() ) );
    SetupAction( "exit",         tr( "E&xit" ),                     QIcon(),                     false,    0,                    tr( "Exits Sudoku" ),                               this,  SLOT( close() ) );
    SetupAction( "undo",         tr( "&Undo" ),                     QIcon( ":/data/undo.png" ),  false,    QKeySequence::Undo,   tr( "Reverts the last change" ),                    this,  SLOT( SudokuUndo() ) );
    SetupAction( "redo",         tr( "&Redo" ),                     QIcon( ":/data/redo.png" ),  false,    QKeySequence::Redo,   tr( "Restores the last undone change" ),            this,  SLOT( SudokuRedo() ) );
    SetupAction( "showallowed",  tr( "Show allowed &numbers" ),     QIcon(),                     true,     0,                    tr( "Shows allowed numbers" ),                      this,  SLOT( SudokuToggleAllowedNumbers() ) );
    SetupAction( "showsimplify", tr( "Simplif&y allowed numbers" ), QIcon(),                     true,     0,                    tr( "Simplifies allowed numbers" ),                 this,  SLOT( SudokuToggleSimplifiedAllowedNumbers() ) );
    SetupAction( "showtips",     tr( "Show solution &tips" ),       QIcon(),                     true,     0,                    tr( "Shows solution tips" ),                        this,  SLOT( SudokuToggleSolutionTips() ) );
    SetupAction( "solveone",     tr( "Solve &one" ),                QIcon(),                     false,    0,                    tr( "Solves one field" ),                           this,  SLOT( SudokuSolveOne() ) );
    SetupAction( "solveall",     tr( "Solve &all" ),                QIcon(),                     false,    0,                    tr( "Solves the complete sudoku" ),                 this,  SLOT( SudokuSolveAll() ) );
    SetupAction( "about",        tr( "&About Sudoku..." ),          QIcon(),                     false,    0,                    tr( "Shows information about Sudoku" ),             this,  SLOT( About() ) );
    SetupAction( "aboutqt",      tr( "About &Qt..." ),              QIcon( ":/data/qt.png" ),    false,    0,                    tr( "Shows information about Qt" ),                 this,  SLOT( AboutQt() ) );
}

void MainWindow::InitMenus()
{
    // Create menus
    //         key         Title
    SetupMenu( "file",     tr( "&File" ) );
    SetupMenu( "recent",   tr( "&Recent files" ) );
    SetupMenu( "edit",     tr( "&Edit" ) );
    SetupMenu( "view",     tr( "&View" ) );
    SetupMenu( "help",     tr( "&Help" ) );

    // Menu for toolbars and dock widgets
    mpMenuToolbars = QMainWindow::createPopupMenu();
    if( mpMenuToolbars ) mpMenuToolbars -> setTitle( tr( "Toolbars" ) );

    // File
    mMenus[ "file" ] -> addAction( mActions[ "generate" ] );
    mMenus[ "file" ] -> addAction( mActions[ "restart" ] );
    mMenus[ "file" ] -> addSeparator();
    mMenus[ "file" ] -> addAction( mActions[ "open" ] );
    mMenus[ "file" ] -> addMenu( mMenus[ "recent" ] );
    mMenus[ "file" ] -> addSeparator();
    mMenus[ "file" ] -> addAction( mActions[ "save" ] );
    mMenus[ "file" ] -> addAction( mActions[ "saveas" ] );
    mMenus[ "file" ] -> addSeparator();
    //mMenus[ "file" ] -> addAction( mActions[ "export" ] );
    //mMenus[ "file" ] -> addSeparator();
    mMenus[ "file" ] -> addAction( mActions[ "print" ] );
    mMenus[ "file" ] -> addAction( mActions[ "printsetup" ] );
    mMenus[ "file" ] -> addSeparator();
    mMenus[ "file" ] -> addAction( mActions[ "exit" ] );

    // Edit
    mMenus[ "edit" ] -> addAction( mActions[ "undo" ] );
    mMenus[ "edit" ] -> addAction( mActions[ "redo" ] );
    mMenus[ "edit" ] -> addSeparator();
    mMenus[ "edit" ] -> addAction( mActions[ "solveone" ] );
    mMenus[ "edit" ] -> addAction( mActions[ "solveall" ] );
    //mMenus[ "edit" ] -> addSeparator();
    //mMenus[ "edit" ] -> addAction( mActions[ "editmode" ] );

    // View
    mMenus[ "view" ] -> addMenu( mpMenuToolbars );
    mMenus[ "view" ] -> addSeparator();
    mMenus[ "view" ] -> addAction( mActions[ "showallowed" ] );
    mMenus[ "view" ] -> addAction( mActions[ "showsimplify" ] );
    mMenus[ "view" ] -> addAction( mActions[ "showtips" ] );

    // Help
    mMenus[ "help" ] -> addAction( mActions[ "about" ] );
    mMenus[ "help" ] -> addAction( mActions[ "aboutqt" ] );

    // Connect some menu items
    connect( mMenus[ "recent" ], SIGNAL( aboutToShow() ), this, SLOT( UpdateMenuRecent() ) );
}

void MainWindow::InitToolBars()
{
    // Create toolbars
    //            key         Title
    SetupToolBar( "standard", tr( "Standard" ) );
    SetupToolBar( "edit",     tr( "Edit" ) );

    // Standard
    mToolBars[ "standard" ] -> addAction( mActions[ "generate" ] );
    mToolBars[ "standard" ] -> addSeparator();
    mToolBars[ "standard" ] -> addAction( mActions[ "open" ] );
    mToolBars[ "standard" ] -> addAction( mActions[ "save" ] );
    mToolBars[ "standard" ] -> addSeparator();
    mToolBars[ "standard" ] -> addAction( mActions[ "print" ] );

    // Edit
    mToolBars[ "edit" ] -> addAction( mActions[ "undo" ] );
    mToolBars[ "edit" ] -> addAction( mActions[ "redo" ] );
}

void MainWindow::InitDockWidgets()
{
    // Create dock widgets
    //               key         Title             widget
    SetupDockWidget( "generate", tr( "Generate" ), new GenerateDockWidget( mpGridView, this ) );
    SetupDockWidget( "solve",    tr( "Solve" ),    new SolveDockWidget( mpGridView, this ) );
    SetupDockWidget( "analyse",  tr( "Analyse" ),  new AnalyseDockWidget( mpGridView, this ) );
}

void MainWindow::InitWidgets()
{
    // Configure widgets
    mpMenuBar -> addMenu( mMenus[ "file" ] );
    mpMenuBar -> addMenu( mMenus[ "edit" ] );
    mpMenuBar -> addMenu( mMenus[ "view" ] );
    mpMenuBar -> addMenu( mMenus[ "help" ] );

    connect( &mRecentMapper, SIGNAL( mapped( QString const & ) ), this, SLOT( SudokuOpen( QString const & ) ) );

    // Add widgets to main window
    setMenuBar( mpMenuBar );
    setCentralWidget( mpGridView );
    statusBar() -> addPermanentWidget( mpLabelProgress );

    connect( mpGridView, SIGNAL( Modified() ),                 this,        SLOT( UpdateState() ) );
    connect( mpGridView, SIGNAL( Message( QString const & ) ), statusBar(), SLOT( showMessage( QString const & ) ) );
    UpdateState();
}

void MainWindow::SetupAction(
    QString const       &rKey,
    QString const       &rText,
    QIcon const         &rIcon,
    bool const          isCheckable,
    QKeySequence const  &rShortcut,
    QString const       &rStatusTip,
    QObject const       *pTarget,
    char const          *pMethod
)
{
    // Setup action
    QAction  *p_action = new QAction( rIcon, rText, this );
    p_action -> setCheckable( isCheckable );
    p_action -> setShortcut( rShortcut );
    p_action -> setStatusTip( rStatusTip );

    connect( p_action, SIGNAL( triggered() ), pTarget, pMethod );

    // Add action to list
    mActions[ rKey ] = p_action;
}

void MainWindow::SetupMenu(
    QString const   &rKey,
    QString const   &rTitle
)
{
    // Setup menu
    QMenu  *p_menu = new QMenu( rTitle, this );

    // Add menu to list
    mMenus[ rKey ] = p_menu;
}

void MainWindow::SetupToolBar(
    QString const   &rKey,
    QString const   &rTitle
)
{
    // Setup toolbar
    QToolBar  *p_toolbar = new QToolBar( rTitle, this );
    p_toolbar -> setObjectName( rKey );
    p_toolbar -> setFloatable( false );
    addToolBar( p_toolbar );

    // Add toolbar to list
    mToolBars[ rKey ] = p_toolbar;
}

void MainWindow::SetupDockWidget(
    QString const   &rKey,
    QString const   &rTitle,
    QWidget         *pWidget
)
{
    // Setup dock widget
    QDockWidget  *p_dock = new QDockWidget( rTitle, this );
    p_dock -> setObjectName( rKey );
    p_dock -> setFeatures( QDockWidget::DockWidgetMovable | QDockWidget::DockWidgetClosable );
    p_dock -> setAllowedAreas( Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea );
    p_dock -> setWidget( pWidget );
    addDockWidget( Qt::LeftDockWidgetArea, p_dock );

    // Add dock widget to list
    mDockWidgets[ rKey ] = p_dock;
}

bool MainWindow::SaveDoc(
    QDomDocument const  &rDoc,
    QString const       &rFileName,
    QString             * const pError
) const
{
    QString  error;
    bool  ok = false;
    QString const  path = QFileInfo( rFileName ) . absolutePath();
    if( QDir() . mkpath( path ) ) {
        QFile  file( rFileName );
        if( file . open( QIODevice::WriteOnly | QIODevice::Truncate ) ) {
            ok = true;
            ok = file . write( "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" ) && ok;
            ok = file . write( rDoc . toByteArray( 4 ) ) && ok;
            file . close();
            if( !ok ) {
                // File write error
                error = tr( "Could not write to file '%1'!\nError: %2" ) . arg( rFileName ) . arg( file . errorString() );
            }
        } else {
            // File create error
            error = tr( "Could not create file '%1'!\nError: %2" ) . arg( rFileName ) . arg( file . errorString() );
        }
    } else {
        // Path create error
        error = tr( "Could not create path '%1'!" ) . arg( path );
    }

    if( pError ) *pError = error;
    return ok;
}

bool MainWindow::LoadDoc(
    QDomDocument   &rDoc,
    QString const  &rFileName,
    QString        * const pError
) const
{
    QString  error;
    bool  ok = false;
    QFile  file( rFileName );
    if( file . open( QIODevice::ReadOnly ) ) {
        // Parse file
        QString  error_text;
        int      error_line = 0, error_col = 0;
        if( rDoc . setContent( &file, &error_text, &error_line, &error_col ) ) {
            ok = true;
        } else {
            // Parsing error
            error = tr( "Error in file '%1':\n%2 (%3, %4)!" ) . arg( rFileName ) . arg( error_text ) . arg( error_line ) . arg( error_col );
        }
        file . close();
    } else {
        // File open error
        error = tr( "Could not read from file '%1'!\nError: %2" ) . arg( rFileName ) . arg( file . errorString() );
    }

    if( pError ) *pError = error;
    return ok;
}

bool MainWindow::LoadConfig()
{
    // Only read config if existing
    QString const  filename = ConfigPath() + "/config.xml";
    bool  ok = false;
    if( QFile::exists( filename ) ) {
        QDomDocument  doc;
        QString  error;
        if( LoadDoc( doc, filename, &error ) ) {
            QDomElement  root = doc . documentElement();
            if( ( root . tagName() == "sudoku" ) && Load( root ) ) {
                ok = true;
            } else {
                // Invalid config file
                QMessageBox::critical( this, tr( "Load error" ), tr( "Config file '%1' is invalid!" ) . arg( filename ), QMessageBox::Ok, 0, 0 );
            }
        } else {
            // Loading failed
            QMessageBox::critical( this, tr( "Load error" ), tr( "Error while loading config '%1':\n%2" ) . arg( filename ) . arg( error ), QMessageBox::Ok, 0, 0 );
        }
    } else {
        // A non existing config file is not an error
        ok = true;
    }

    return ok;
}

bool MainWindow::Load(
    QDomElement const  &rParent
)
{
    // Read config
    QDomElement  config = rParent . firstChildElement( "config" );
    if( config . isNull() ) {
        // The config tag is mandatory
        return false;
    }

    mRecentFiles . clear();

    // Parse all elements
    for( QDomNode  node = config . firstChild(); ! node . isNull(); node = node . nextSibling() ) {
        if( node . isElement() ) {
            QDomElement  element = node . toElement();
            if( element . tagName() == "state" ) {
                // Window state
                mState = qUncompress( QByteArray::fromBase64( element . text() . toLatin1() ) );
                mStateVersion = element . attribute( "version", "0" ) . toInt();
            } else if( element . tagName() == "geometry" ) {
                int   geo_x = element . attribute( "x", "100" ) . toInt();
                int   geo_y = element . attribute( "y", "100" ) . toInt();
                int   geo_width = element . attribute( "width", "300" ) . toInt();
                int   geo_height = element . attribute( "height", "300" ) . toInt();
                bool  geo_maximized = element . attribute( "maximized", "0" ) . toInt();
                move( geo_x, geo_y );
                resize( geo_width, geo_height );

                if( geo_maximized ) setWindowState( windowState() | Qt::WindowMaximized );
                else                setWindowState( windowState() & ~Qt::WindowMaximized );
            } else if( element . tagName() == "recentfile" ) {
                // Recent file entry
                mRecentFiles . append( element . text() );
            }
        }
    }

    return true;
}

bool MainWindow::SaveConfig()
{
    QDomDocument  doc;
    QDomElement   root = doc . createElement( "sudoku" );
    doc . appendChild( root );
    bool  ok = false;
    if( Save( root ) ) {
        QString const  filename = ConfigPath() + "/config.xml";
        QString  error;
        if( SaveDoc( doc, filename, &error ) ) {
            ok = true;
        } else {
            // Saving failed
            QMessageBox::critical( this, tr( "Save error" ), tr( "Error while saving config '%1':\n%2" ) . arg( filename ) . arg( error ), QMessageBox::Ok, 0, 0 );
        }
    } else {
        // Build XML failed
        QMessageBox::critical( this, tr( "Save error" ), tr( "Could not save config!" ), QMessageBox::Ok, 0, 0 );
    }

    return ok;
}

bool MainWindow::Save(
    QDomElement  &rParent
) const
{
    QDomDocument  doc = rParent . ownerDocument();

    if( doc . isNull() ) {
        // A document is needed to create sub tags
        return false;
    }

    // Config main structure
    QDomElement   config = doc . createElement( "config" );
    rParent . appendChild( config );

    // Main window state
    int const  state_version = 1;
    QDomElement  state = doc . createElement( "state" );
    state . setAttribute( "version", state_version );
    QDomText  state_text = doc . createTextNode( qCompress( saveState( state_version ), 9 ) . toBase64() );
    state . appendChild( state_text );
    config . appendChild( state );

    // Position
    bool const  geo_maximized = isMaximized();
    QRect const  geo_rect = normalGeometry();
    QDomElement  geo = doc . createElement( "geometry" );
    geo . setAttribute( "x", geo_rect . x() );
    geo . setAttribute( "y", geo_rect . y() );
    geo . setAttribute( "width", geo_rect . width() );
    geo . setAttribute( "height", geo_rect . height() );
    geo . setAttribute( "maximized", geo_maximized );
    config . appendChild( geo );

    // Recent files
    QString  recent_file;
    Q_FOREACH( recent_file, mRecentFiles ) {
        QDomElement  recent = doc . createElement( "recentfile" );
        QDomText  recent_text = doc . createTextNode( recent_file );
        recent . appendChild( recent_text );
        config . appendChild( recent );
    }

    return true;
}

void MainWindow::About()
{
    char const  * const p_year = strrchr( __DATE__, ' ' );
    int const  copy_year1 = 2007, copy_year2 = p_year ? atoi( p_year ) : 2013;
    QString const  message = "<h3><nobr>" + tr( "Sudoku - Create and play sudoku" ) + "</nobr></h3>" + tr( "<p>Version %1</p><p><nobr>Copyright &copy; %2-%3 Sven Heinzel</nobr><br>Visit the <a href='http://www.sternenwanderer.net/project/sudoku.shtml'>Sudoku homepage</a></p><p>Sudoku comes with ABSOLUTELY NO WARRANTY. This is free software, and you are welcome to redistribute it under certain conditions. See <a href='GPL.html'>GPL.html</a> for details.</p>" ) . arg( QString::fromStdString( Sudoku::Version::String() ) ) . arg( copy_year1 ) . arg( copy_year2 );
    QMessageBox::about( this, tr( "About Sudoku" ), message );
}

void MainWindow::AboutQt()
{
    QMessageBox::aboutQt( this );
}

void MainWindow::UpdateMenuRecent()
{
    // Remove all recent list actions
    QAction  *p_action = 0;
    foreach( p_action, mRecentList ) delete p_action;
    mRecentList . clear();

    // Create entries
    int  i = 0;
    for( QStringList::const_iterator  it = mRecentFiles . constBegin(); it != mRecentFiles . constEnd(); ++it, ++i )
    {
        QString const  file = *it;

        // Create entry
        QString  title;
        if( i < 9 ) title = QString( "&%1 %2" ) . arg( i ) . arg( QDir::toNativeSeparators( file ) );
        else        title = QString( "%1 %2" ) . arg( i ) . arg( QDir::toNativeSeparators( file ) );

        // Add action
        p_action = mMenus[ "recent" ] -> addAction( title, &mRecentMapper, SLOT( map() ) );
        if( p_action ) {
            mRecentMapper . setMapping( p_action, file );

            // Append to list
            mRecentList . append( p_action );

            // Set first one as active entry
            if( i == 1 ) mMenus[ "recent" ] -> setActiveAction( p_action );
        }
    }

    // Add default entry if list is empty
    if( i == 0 ) {
        p_action = mMenus[ "recent" ] -> addAction( tr( "<Empty>" ) );
        if( p_action ) {
            // Disable action
            p_action -> setEnabled( false );

            // Append to list
            mRecentList . append( p_action );
        }
    }
}

void MainWindow::UpdateState()
{
    mActions[ "undo" ] -> setEnabled( mpGridView -> CanUndo() );
    mActions[ "redo" ] -> setEnabled( mpGridView -> CanRedo() );
    mActions[ "showallowed" ] -> setChecked( mpGridView -> GetDisplayOptions() . IsVisible( DisplayOptions::ShowAllowed ) );
    mActions[ "showsimplify" ] -> setChecked( mpGridView -> GetDisplayOptions() . IsVisible( DisplayOptions::ShowSimplifiedAllowed ) );
    mActions[ "showsimplify" ] -> setEnabled( mpGridView -> GetDisplayOptions() . IsVisible( DisplayOptions::ShowAllowed ) );
    mActions[ "showtips" ] -> setChecked( mpGridView -> GetDisplayOptions() . IsVisible( DisplayOptions::ShowTips ) );

    int const  filled = mpGridView -> GetFilledFields();
    int const  total = mpGridView -> GetTotalFields();
    int const  progress = 100 * filled / total;
    mpLabelProgress -> setText( tr( "%1 of %2 fields done (%3%)" ) . arg( filled ) . arg( total ) . arg( progress) );
}

void MainWindow::AddRecentFile(
    QString const   &rFile
)
{
    if( ! rFile . isEmpty() ) {
        // Move entry to start of list
        mRecentFiles . removeAll( rFile );
        mRecentFiles . prepend( rFile );

        // Truncate
        mRecentFiles = mRecentFiles . mid( 0, 4 );
    }
}

void MainWindow::SudokuRestart()
{
    // Restart
    mpGridView -> Restart();
}

void MainWindow::SudokuGenerate()
{
    // Open generate dialog
    GenerateDialog  dialog( mGenerateOptions, this );

    // If dialog was accepted generate sudoku
    if( dialog . exec() == QDialog::Accepted ) {
        mGenerateOptions = dialog . GetGenerateOptions();
        SudokuGenerate( mGenerateOptions );
    }
}

void MainWindow::SudokuGenerate(
    Sudoku::GenerateOptions const  &rOptions
)
{
    // Generate sudoku
    mpGridView -> Generate( rOptions );
    mFileName = QString();
}

void MainWindow::SudokuOpen(
    QString const   &rFile
)
{
    // Open file selection when no name is given
    QString  filename = rFile;
    if( filename . isEmpty() ) {
        QString const  path = mFileName . isEmpty() ? mLastPath : mFileName;
        filename = QFileDialog::getOpenFileName( this, tr( "Load" ), path, tr( "Sudoku grid (*.sf *.sg)" ) );
    }

    // If we have a file name now open it
    if( ! filename . isEmpty() ) {
        SudokuLoad( filename );
    }
}

void MainWindow::SudokuSave()
{
    if( mFileName . isEmpty() ) {
        SudokuSaveAs();
    } else {
        SudokuSave( mFileName );
    }
}

void MainWindow::SudokuSaveAs()
{
    // Open file selection
    QStringList  extensions;
    extensions << tr( "Sudoku games (*.sg)" ) << tr( "Sudoku fixed fields only (*.sf)" );
    QString  ext_sel;
    QString const  path = mFileName . isEmpty() ? mLastPath : mFileName;
    QString  filename = QFileDialog::getSaveFileName( this, tr( "Save" ), path, extensions . join( "\n" ), &ext_sel );

    // If the user selcted a file try to open it
    if( ! filename . isEmpty() ) {
        // Get extension
        int  sel = extensions . indexOf( ext_sel );
        QString  extension;
        if( sel == 0 )      extension = "sg";
        else if( sel == 1 ) extension = "sf";

        if( extension == "sf" ) {
            QMessageBox::information( this, tr( "Save grid" ), tr( "This will only save the fixed fields." ), QMessageBox::Ok, 0, 0 );
        }

        // Add default extension
        if( QFileInfo( filename ) . suffix() . toLower() != extension ) {
            filename = QFileInfo( filename ) . completeBaseName() + "." + extension;
        }

        SudokuSave( filename );
    }
}

bool MainWindow::SudokuSave(
    QString const   &rFileName
)
{
    // Store sudoku to file
    bool const  game_mode = ( QFileInfo( rFileName ) . suffix() == "sg" );
    QDomDocument  doc;
    QDomElement   root = doc . createElement( "SUDOKU" );
    doc . appendChild( root );
    bool  ok = false;
    if( mpGridView -> Save( root, game_mode ) ) {
        QString  error;
        if( SaveDoc( doc, rFileName, &error ) ) {
            ok = true;
            mLastPath = QFileInfo( rFileName ) . absolutePath();
            mFileName = rFileName;
            AddRecentFile( rFileName );
        } else {
            // Saving failed
            QMessageBox::critical( this, tr( "Save error" ), tr( "Error while saving sudoku '%1':\n%2" ) . arg( rFileName ) . arg( error ), QMessageBox::Ok, 0, 0 );
        }
    } else {
        // Build XML failed
        QMessageBox::critical( this, tr( "Save error" ), tr( "Could not save sudoku!" ), QMessageBox::Ok, 0, 0 );
    }

    return ok;
}

bool MainWindow::SudokuLoad(
    QString const   &rFileName
)
{
    // Only read config if existing
    bool  ok = false;
    QDomDocument  doc;
    QString  error;
    if( LoadDoc( doc, rFileName, &error ) ) {
        QDomElement  root = doc . documentElement();
        if( ( root . tagName() == "SUDOKU" ) && mpGridView -> Load( root ) ) {
            ok = true;
            mLastPath = QFileInfo( rFileName ) . absolutePath();
            mFileName = rFileName;
            AddRecentFile( rFileName );
        } else {
            // Invalid sudoku file
            QMessageBox::critical( this, tr( "Load error" ), tr( "Sudoku file '%1' is invalid!" ) . arg( rFileName ), QMessageBox::Ok, 0, 0 );
        }
    } else {
        // Loading failed
        QMessageBox::critical( this, tr( "Load error" ), tr( "Error while loading sudoku '%1':\n%2" ) . arg( rFileName ) . arg( error ), QMessageBox::Ok, 0, 0 );
    }

    return ok;
}

void MainWindow::SudokuExport()
{
    // Export
}

void MainWindow::SudokuPrint()
{
    // Show preview
    QPrintPreviewDialog  dialog( &mPrinter, this );
    connect( &dialog, SIGNAL( paintRequested( QPrinter* ) ), this, SLOT( DrawPrint( QPrinter* ) ) );

    // If the user accepted the dialog print
    if( dialog . exec() == QDialog::Accepted ) {
        DrawPrint( &mPrinter );
    }
}

void MainWindow::SudokuPrinterSetup()
{
    QPrintDialog  dialog( &mPrinter, this );
    dialog . exec();
}

void MainWindow::SudokuUndo()
{
    // Undo
    mpGridView -> Undo();
}

void MainWindow::SudokuRedo()
{
    // Redo
    mpGridView -> Redo();
}

void MainWindow::SudokuToggleAllowedNumbers()
{
    // Toggle allowed numbers
    DisplayOptions  options = mpGridView -> GetDisplayOptions();
    options . ToggleVisible( DisplayOptions::ShowAllowed );
    mpGridView -> SetDisplayOptions( options );
}

void MainWindow::SudokuToggleSimplifiedAllowedNumbers()
{
    // Toggle simpliefied allowed numbers
    DisplayOptions  options = mpGridView -> GetDisplayOptions();
    options . ToggleVisible( DisplayOptions::ShowSimplifiedAllowed );
    mpGridView -> SetDisplayOptions( options );
}

void MainWindow::SudokuToggleSolutionTips()
{
    // Toggle solution tips
    DisplayOptions  options = mpGridView -> GetDisplayOptions();
    options . ToggleVisible( DisplayOptions::ShowTips );
    mpGridView -> SetDisplayOptions( options );
}

void MainWindow::SudokuSolveOne()
{
    // Solve one
    mpGridView -> SolveOne();
}

void MainWindow::SudokuSolveAll()
{
    // Solve all
    mpGridView -> SolveAll();
}

void MainWindow::DrawPrint(
    QPrinter  *pPrinter
)
{
    QPainter  painter;
    if( painter . begin( pPrinter ) ) {
        // Draw the sudoku
        mpGridView -> DrawSudoku( painter, 100.0, DisplayOptions() );
        painter . end();
    }
}
