/*
sudoku - Create and play sudoku

Copyright (C) 2007 Sven Heinzel <Sven.Heinzel@gmx.net>
 
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#pragma once

#ifndef SUDOKU_ALLOWED_H
#define SUDOKU_ALLOWED_H

// No use of class AllowedCount for now
#ifndef SUDOKU_ENABLE_ALLOWED_COUNT
#define SUDOKU_ENABLE_ALLOWED_COUNT 0
#endif

#include "sudoku/sudoku.h"

namespace Sudoku {

#if SUDOKU_ENABLE_ALLOWED_COUNT
class AllowedCount
{
public:
    AllowedCount()
    {
        Reset();
    }

    AllowedCount( AllowedCount const  &rOther )
    {
        Copy( rOther );
    }

    AllowedCount( uint64_t const  mask )
    {
        Reset();
        int  i = 0;
        for( uint64_t  m = mask; m; m >>= 1, ++i ) {
            int const  val = m & 1;
            mTotal += val;
            mCount[ i ] = val;
        }
    }

    ~AllowedCount()
    {
    }

    inline void Reset()
    {
        for( int  i = 0; i < 64; ++i ) {
            mCount[ i ] = 0;
        }
        mTotal = 0;
    }

    inline int GetCount( int const  num ) const
    {
        return ( 0 < num && num <= 64 ) ? mCount[ num - 1 ] : 0;
    }

    inline int GetTotalCount() const
    {
        return mTotal;
    }

    inline bool operator==( AllowedCount const  &rOther )
    {
        if( mTotal != rOther . mTotal ) return false;
        for( int  i = 0; i < 64; ++i ) {
            bool const  equal = mCount[ i ] == rOther . mCount[ i ];
            if( ! equal ) return false;
        }
        return true;
    }

    inline bool operator!=( AllowedCount const  &rOther )
    {
        return ! ( *this == rOther );
    }

    inline AllowedCount & operator=( AllowedCount const  &rOther )
    {
        if( this != &rOther ) {
            Copy( rOther );
        }
        return *this;
    }

    inline AllowedCount operator+( AllowedCount const  &rOther ) const
    {
        AllowedCount  count( *this );
        Add( count, rOther );
        return count;
    }

    inline AllowedCount & operator+=( AllowedCount const  &rOther )
    {
        Add( *this, rOther );
        return *this;
    }

private:
    static inline void Add( AllowedCount  &rDest, AllowedCount const  &rSrc )
    {
        for( int  i = 0; i < 64; ++i ) {
            rDest . mCount[ i ] += rSrc . mCount[ i ];
        }
        rDest . mTotal += rSrc . mTotal;
    }

    inline void Copy( AllowedCount const  &rOther )
    {
        for( int  i = 0; i < 64; ++i ) {
            mCount[ i ] = rOther . mCount[ i ];
        }
        mTotal = rOther . mTotal;
    }

    int  mCount[ 64 ];
    int  mTotal;
};
#endif

class Allowed
{
public:
    Allowed( int const  count = 9 )
        : mCount( std::max( 4, std::min( count, 64 ) ) )
        , mMask( mCount < 64 ? ( 1ull << mCount ) - 1ull : ~0ull )
    {
        Reset();
    }

    Allowed( Allowed const  &rOther )
    {
        Copy( rOther );
    }

    ~Allowed()
    {
    }

    void Reset( bool const  fieldOnly = false )
    {
        if( ! fieldOnly ) {
            // Clear standard masks (only if not fieldOnly)
            for( int  i = 0; i < mCount; ++i ) {
                mRow[ i ]    = mMask;
                mColumn[ i ] = mMask;
                mBox[ i ]    = mMask;
            }
        }
        // Clear extra field masks (always)
        for( int  y = 0; y < mCount; ++y ) {
            for( int  x = 0; x < mCount; ++x ) {
                mField[ y ][ x ] = mMask;
            }
        }
    }

    inline void MarkUsed( int const  x, int const  y, int const  b, int const  num )
    {
        if( num <= 0 || ! CheckRangeXYB( x, y, b ) ) return;
        // Clear bit
        uint64_t const  mask = ~NumberMask( num );
        mField[ y ][ x ] = 0; // This field is totally occupied
        mColumn[ x ] &= mask;
        mRow[ y ]    &= mask;
        mBox[ b ]    &= mask;
    }

    inline void MarkUnused( int const  x, int const  y, int const  b, int const  num )
    {
        if( num <= 0 || ! CheckRangeXYB( x, y, b ) ) return;
        // Set bit
        uint64_t const  mask = NumberMask( num ) & mMask;
        mField[ y ][ x ] = mMask; // This field is totally unoccupied
        mColumn[ x ] |= mask;
        mRow[ y ]    |= mask;
        mBox[ b ]    |= mask;
    }

    inline void SetFieldMask( int const  x, int const  y, uint64_t const  mask )
    {
        if( ! CheckRangeXY( x, y ) ) return;
        mField[ y ][ x ] = mask & mMask;
    }

    inline uint64_t GetAllowedMask( int const  x, int const  y, int const  b ) const
    {
        return CheckRangeXYB( x, y, b ) ? AllowedMask( x, y, b ) : 0;
    }

    inline bool IsNumberAllowed( int const  x, int const  y, int const  b, int const  num ) const
    {
        return CheckRangeXYB( x, y, b ) ? ( num == 0 ) || ( AllowedMask( x, y, b ) & NumberMask( num ) ) : false;
    }

    static inline int CountAllowed( uint64_t const  mask )
    {
        uint64_t  count = mask - ( ( mask >> 1 ) & 0x5555555555555555ull );
        count = ( count & 0x3333333333333333ull ) + ( ( count >> 2 ) & 0x3333333333333333ull );
        count = ( count + ( count >> 4 ) ) & 0x0f0f0f0f0f0f0f0full;
        return static_cast< int >( ( count * 0x0101010101010101ull ) >> 56 );
    }

    static inline bool IsOneAllowed( uint64_t const  mask )
    {
        // Clear LSB and check if there are no bits left
        return ( mask & ( mask - 1 ) ) == 0;
    }

    static inline uint64_t NumberMask( int const  num )
    {
        return ( 0 < num && num <= 64 ) ? ( 1 << ( num - 1 ) ) : 0;
    }

    inline bool operator==( Allowed const  &rOther ) const
    {
        if( mCount != rOther . mCount ) return false;
        for( int  i = 0; i < mCount; ++i ) {
            bool const  equal =    mRow[ i ]    == rOther . mRow[ i ]
                                && mColumn[ i ] == rOther . mColumn[ i ]
                                && mBox[ i ]    == rOther . mBox[ i ];
            if( ! equal ) return false;
        }
        for( int  y = 0; y < mCount; ++y ) {
            for( int  x = 0; x < mCount; ++x ) {
                bool const  equal = mField[ y ][ x ] == rOther . mField[ y ][ x ];
                if( ! equal ) return false;
            }
        }
        return true;
    }

    inline bool operator!=( Allowed const  &rOther ) const
    {
        return ! ( *this == rOther );
    }

    inline Allowed & operator=( Allowed const  &rOther )
    {
        if( this != &rOther ) {
            Copy( rOther );
        }
        return *this;
    }

private:
    inline bool CheckRangeXY( int const  x, int const  y ) const
    {
        return ( 0 <= x && x < mCount ) && ( 0 <= y && y < mCount );
    }

    inline bool CheckRangeXYB( int const  x, int const  y, int const  b ) const
    {
        return ( 0 <= x && x < mCount ) && ( 0 <= y && y < mCount ) && ( 0 <= b && b < mCount );
    }

    inline uint64_t AllowedMask( int const  x, int const  y, int const  b ) const
    {
        return mColumn[ x ] & mRow[ y ] & mBox[ b ] & mField[ y ][ x ];
    }

    inline void Copy( Allowed const  &rOther )
    {
        mCount = rOther . mCount;
        mMask = rOther . mMask;
        for( int  i = 0; i < mCount; ++i ) {
            mColumn[ i ] = rOther . mColumn[ i ];
            mRow[ i ]    = rOther . mRow[ i ];
            mBox[ i ]    = rOther . mBox[ i ];
        }
        for( int  y = 0; y < mCount; ++y ) {
            for( int  x = 0; x < mCount; ++x ) {
                mField[ y ][ x ] = rOther . mField[ y ][ x ];
            }
        }
    }

    int  mCount;
    uint64_t  mMask;
    uint64_t  mColumn[ 64 ];
    uint64_t  mBox[ 64 ];
    uint64_t  mRow[ 64 ];
    uint64_t  mField[ 64 ][ 64 ];
};

}

#endif
