/*
sudoku - Create and play sudoku

Copyright (C) 2007 Sven Heinzel <Sven.Heinzel@gmx.net>
 
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#pragma once

#ifndef SUDOKU_GRID_H
#define SUDOKU_GRID_H

#include "sudoku/sudoku.h"
#include "sudoku/Allowed.h"
#include "sudoku/Layout.h"

namespace Sudoku {

class Grid
{
public:
    Grid( int const  count = 9, bool const  randomLayout = false )
    {
        Resize( count, randomLayout );
    }

    Grid( Grid const  &rOther )
    {
        Copy( rOther );
    }

    ~Grid()
    {
    }

    void Reset( bool  resetFixed = true )
    {
        if( resetFixed ) {
            // Clear everything
            for( int  y = 0; y < mCount; ++y ) {
                for( int  x = 0; x < mCount; ++x ) {
                    mNumber[ y ][ x ] = 0;
                    mFixed[ y ][ x ] = false;
                }
            }
            mAllowed . Reset();
            mFreeFields = GetTotalFields();
        } else {
            // Only clear non fixed fields
            for( int  y = 0; y < mCount; ++y ) {
                for( int  x = 0; x < mCount; ++x ) {
                    if( ! mFixed[ y ][ x ] ) MarkUnused( x, y );
                }
            }
        }
    }

    inline void Resize( int const  count = 9, bool const  randomLayout = false )
    {
        mCount = std::max( 4, std::min( count, 64 ) );
        mLayout . Resize( mCount, randomLayout );
        mAllowed = Allowed( mCount );
        Reset();
    }

    inline bool SetNumber( int const  x, int const  y, int const  num )
    {
        if( IsFixed( x, y ) || ! IsNumberAllowed( x, y, num ) ) return false;

        MarkUnused( x, y );
        MarkUsed( x, y, num );

        return true;
    }

    inline void SetFixed( int const  x, int const  y, bool const  fixed = true )
    {
        if( ! CheckRange( x, y ) ) return;
        mFixed[ y ][ x ] = fixed;
    }

    inline bool IsFixed( int const  x, int const  y ) const
    {
        return CheckRange( x, y ) ? mFixed[ y ][ x ] : false;
    }

    inline Allowed const & GetAllowed() const
    {
        return mAllowed;
    }

    inline Layout const & GetLayout() const
    {
        return mLayout;
    }

    inline int BoxIndex( int const  x, int const  y ) const
    {
        return mLayout . BoxIndex( x, y );
    }

    inline uint64_t GetAllowedMask( int const  x, int const  y ) const
    {
        // Range check is done inside Allowed
        return mAllowed . GetAllowedMask( x, y, BoxIndex( x, y ) );
    }

    inline bool IsNumberAllowed( int const  x, int const  y, int const  num ) const
    {
        // Range check is done inside Allowed
        return mAllowed . IsNumberAllowed( x, y, BoxIndex( x, y ), num );
    }

    inline int GetNumber( int const  x, int const  y ) const
    {
        return CheckRange( x, y ) ? mNumber[ y ][ x ] : -1;
    }

    inline bool IsValid() const
    {
        for( int  y = 0; y < mCount; ++y ) {
            for( int  x = 0; x < mCount; ++x ) {
                if( mNumber[ y ][ x ] == 0 && mAllowed . GetAllowedMask( x, y, BoxIndex( x, y ) ) == 0 ) {
                    // Nothing allowed in an empty field so there is no solution
                    return false;
                }
            }
        }
        return true;
    }

    inline bool IsComplete() const
    {
        return mFreeFields == 0;
    }

    inline int GetFreeFields() const
    {
        return mFreeFields;
    }

    inline int GetFilledFields() const
    {
        return GetTotalFields() - mFreeFields;
    }

    inline int GetTotalFields() const
    {
        return mCount * mCount;
    }

    inline int GetCount() const
    {
        return mCount;
    }

#if 0
    inline void Dump() const
    {
        // NEEDS TO BE REWRITTEN
        for( int  y = 0; y < 9; ++y ) {
            if( y % 3 == 0 ) std::cout << "+---------+---------+---------+" << std::endl;
            for( int  x = 0; x < 9; ++x ) {
                if( x % 3 == 0 ) std::cout << "|";
                if( IsFixed( x, y ) ) std::cout << "[" << mNumber[ y ][ x ] << "]";
                else std::cout << " " << mNumber[ y ][ x ] << " ";
            }
            std::cout << "|" << std::endl;
        }
        std::cout << "+---------+---------+---------+" << std::endl << std::endl;
    }
#endif

    inline bool operator==( Grid const  &rOther )
    {
        if(    mCount      != rOther . mCount
            || mFreeFields != rOther . mFreeFields
            || mAllowed    != rOther . mAllowed
            || mLayout     != rOther . mLayout ) return false;
        for( int  y = 0; y < mCount; ++y ) {
            for( int  x = 0; x < mCount; ++x ) {
                bool const  equal =    mNumber[ y ][ x ] == rOther . mNumber[ y ][ x ]
                                    && mFixed[ y ][ x ]  == rOther . mFixed[ y ][ x ];
                if( ! equal ) return false;
            }
        }
        return true;
    }

    inline bool operator!=( Grid const  &rOther )
    {
        return ! ( *this == rOther );
    }

    inline Grid & operator=( Grid const  &rOther )
    {
        if( this != &rOther ) {
            Copy( rOther );
        }
        return *this;
    }

private:
    inline bool CheckRange( int const  x, int const  y ) const
    {
        return ( 0 <= x && x < mCount ) && ( 0 <= y && y < mCount );
    }

    inline bool CheckRange( int const  x, int const  y, int const  num ) const
    {
        return ( 0 <= x && x < mCount ) && ( 0 <= y && y < mCount ) && ( 0 <= num && num <= mCount );
    }

    inline void MarkUsed( int const  x, int const  y, int const  num )
    {
        if( num == 0 ) return;
        mNumber[ y ][ x ] = num;
        mAllowed . MarkUsed( x, y, BoxIndex( x, y ), num );
        --mFreeFields;
    }

    inline void MarkUnused( int const  x, int const  y )
    {
        if( mNumber[ y ][ x ] == 0 ) return;
        mAllowed . MarkUnused( x, y, BoxIndex( x, y ), mNumber[ y ][ x ] );
        mNumber[ y ][ x ] = 0;
        ++mFreeFields;
    }

    inline void Copy( Grid const  &rOther )
    {
        mCount      = rOther . mCount;
        mFreeFields = rOther . mFreeFields;
        for( int  y = 0; y < mCount; ++y ) {
            for( int  x = 0; x < mCount; ++x ) {
                mNumber[ y ][ x ] = rOther . mNumber[ y ][ x ];
                mFixed[ y ][ x ]  = rOther . mFixed[ y ][ x ];
            }
        }
        mAllowed = rOther . mAllowed;
        mLayout  = rOther . mLayout;
    }

    int      mCount;
    int      mFreeFields;
    int      mNumber[ 64 ][ 64 ];
    bool     mFixed[ 64 ][ 64 ];
    Allowed  mAllowed;
    Layout   mLayout;
};

}

#endif
