/*
sudoku - Create and play sudoku

Copyright (C) 2007 Sven Heinzel <Sven.Heinzel@gmx.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#pragma once

#ifndef SUDOKU_SOLVER_H
#define SUDOKU_SOLVER_H

#include "sudoku/sudoku.h"
#include "sudoku/SolveOptions.h"
#include "sudoku/Grid.h"

namespace Sudoku {

class SolveResult : public SolveOptions
{
public:
    enum Solution
    {
        NoSolution = 0,
        UniqueSolution,
        MultipleSolutions
    };

    SolveResult()
        : SolveOptions( SolveOptions::none() )
        , mSolution( NoSolution )
    {
    }

    SolveResult( Solution const  solution, SolveOptions const  &rUsedSolveOptions )
        : SolveOptions( rUsedSolveOptions )
        , mSolution( solution )
    {
    }

    ~SolveResult()
    {
    }

    inline void SetSolution( Solution const  solution )
    {
        mSolution = solution;
    }

    inline Solution GetSolution() const
    {
        return mSolution;
    }

private:
    Solution  mSolution;
};

class Solver
{
public:
    Solver()
        : mSolveOptions( SolveOptions::all() )
    {
    }

    Solver( SolveOptions const  &rSolveOptions )
        : mSolveOptions( rSolveOptions )
    {
    }

    ~Solver()
    {
    }

    inline void SetSolveOptions( SolveOptions const  &rSolveOptions )
    {
        mSolveOptions = rSolveOptions;
    }

    inline SolveOptions const & GetSolveOptions() const
    {
        return mSolveOptions;
    }

    bool SolveOne( Grid  &rGrid ) const;
    bool SolveAll( Grid  &rGrid ) const;
    SolveResult Analyse( Grid const  &rGrid ) const;
    Grid FindTips( Grid const  &rGrid, bool const  simplifyAllowed ) const;
    Allowed SimplifyAllowed( Grid const  &rGrid ) const;
    int CheckUnique( Grid const  &rGrid, int const  x, int const  y ) const;

    inline bool operator()( Grid  &rGrid ) const
    {
        return SolveAll( rGrid );
    }

private:
    class Instance
    {
    public:
        Instance( SolveOptions const  &rSolveOptions, Grid  &rGrid )
            : mSolveOptions( rSolveOptions )
            , mSolveResult()
            , mrGrid( rGrid )
            , mAllowed( rGrid . GetAllowed() )
            , mCount( rGrid . GetCount() )
            , mFoundSolutions( 0 )
        {
        }

        ~Instance()
        {
        }

        inline SolveResult GetSolveResult() const
        {
            return mSolveResult;
        }

        inline Allowed const & GetAllowed() const
        {
            return mAllowed;
        }

        bool SimplifyAllowed();
        bool Solve( bool const  checkMultiple );
        bool SolveUniques();
        
        int CheckUnique( int const  x, int const  y ) const;

    private:
        bool BackTrack( bool const  checkMultiple );

        inline bool SetNumber( int const  x, int const  y, int const  num )
        {
            int const  old_num = mrGrid . GetNumber( x, y );
            if( mrGrid . SetNumber( x, y, num ) ) {
                mAllowed . MarkUnused( x, y, mrGrid . BoxIndex( x, y ), old_num );
                mAllowed . MarkUsed( x, y, mrGrid . BoxIndex( x, y ), num );
                return true;
            }
            return false;
        }

        bool CheckPairCol( int const  x, int const  level );
        bool CheckPairRow( int const  y, int const  level );
        bool CheckPairBox( int const  b, int const  level );
        bool CheckPair( int const  level );

        inline bool IsValid() const
        {
            for( int  y = 0; y < mCount; ++y ) {
                for( int  x = 0; x < mCount; ++x ) {
                    if( mrGrid . GetNumber( x, y ) == 0 && mAllowed . GetAllowedMask( x, y, mrGrid . BoxIndex( x, y ) ) == 0 ) {
                        // Nothing allowed in an empty field so there is no solution
                        return false;
                    }
                }
            }
            return true;
        }

        inline uint64_t GetAllowedMask( int const  x, int const  y ) const
        {
            return mAllowed . GetAllowedMask( x, y, mrGrid . BoxIndex( x, y ) );
        }

        inline uint64_t GetAllowedMask( int const  x, int const  y, int const  b ) const
        {
            return mAllowed . GetAllowedMask( x, y, b );
        }

        inline uint64_t UniqueMaskRow( int const  y ) const
        {
            uint64_t  result = 0;
            uint64_t  mask = ~0ull;
            for( int  x = 0; x < mCount; ++x ) {
                // Mask of allowed numbers per field
                uint64_t const  field = mAllowed . GetAllowedMask( x, y, mrGrid . BoxIndex( x, y ) );
                // Mask of numbers that are not unique for this check
                uint64_t const  delta = field & result;
                // Now add unique numbers to result (and mask out all previously found invalid numbers)
                result = ( result ^ field ) & mask;
                // Finally remove non uniques from mask (they will never be unique again)
                mask &= ~delta;
            }
            return result;
        }

        inline uint64_t UniqueMaskColumn( int const  x ) const
        {
            uint64_t  result = 0;
            uint64_t  mask = ~0ull;
            for( int  y = 0; y < mCount; ++y ) {
                // Mask of allowed numbers per field
                uint64_t const  field = mAllowed . GetAllowedMask( x, y, mrGrid . BoxIndex( x, y ) );
                // Mask of numbers that are not unique for this check
                uint64_t const  delta = field & result;
                // Now add unique numbers to result (and mask out all previously found invalid numbers)
                result = ( result ^ field ) & mask;
                // Finally remove non uniques from mask (they will never be unique again)
                mask &= ~delta;
            }
            return result;
        }

        inline uint64_t UniqueMaskBox( int const  b ) const
        {
            uint64_t  result = 0;
            uint64_t  mask = ~0ull;
            for( int  y = 0; y < mCount; ++y ) {
                for( int  x = 0; x < mCount; ++x ) {
                    if( mrGrid . BoxIndex( x, y ) == b ) {
                        // Mask of allowed numbers per field
                        uint64_t const  field = mAllowed . GetAllowedMask( x, y, b );
                        // Mask of numbers that are not unique for this check
                        uint64_t const  delta = field & result;
                        // Now add unique numbers to result (and mask out all previously found invalid numbers)
                        result = ( result ^ field ) & mask;
                        // Finally remove non uniques from mask (they will never be unique again)
                        mask &= ~delta;
                    }
                }
            }
            return result;
        }

        inline void IncreaseMaxPairSolving( int const  level )
        {
            mSolveResult . SetMaxPairSolving( std::max( level, mSolveResult . GetMaxPairSolving() ) );
        }

        inline void IncreaseSolution( SolveResult::Solution const  solution )
        {
            mSolveResult . SetSolution( std::max( solution, mSolveResult . GetSolution() ) );
        }

        SolveOptions  mSolveOptions;
        SolveResult   mSolveResult;
        Grid          &mrGrid;
        Allowed       mAllowed;
        int const     mCount;
        int           mFoundSolutions;
    };

    SolveOptions  mSolveOptions;
};

}

#endif
