/*
sudoku - Create and play sudoku

Copyright (C) 2007 Sven Heinzel <Sven.Heinzel@gmx.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "sudoku/sudoku.h"
#include "sudoku/Grid.h"
#include "sudoku/Solver.h"
#include "sudoku/Generator.h"

namespace Sudoku {

bool Generator::Instance::Generate()
{
    mrGrid . Reset();

    // Fill and remove again
    if( ! FillFields() ) return false;
    bool const  ret = RemoveFields();

    // Always fix all remaining fields
    FixFields();
    return ret;
}

bool Generator::Instance::FillFields( int const  sIx, int const  sIy )
{
    // Check if already completed
    if( mrGrid . IsComplete() ) {
        return true;
    }

    // Now try to find empty fields to fill
    int  start_ix = sIx;
    int  start_iy = sIy;
    for( int  iy = start_iy; iy < mCount; ++iy ) {
        int const  y = mRowOrder[ iy ];
        std::vector< int >  &r_col_order = mColOrder[ iy ];
        for( int  ix = start_ix; ix < mCount; ++ix ) {
            int const  x = r_col_order[ ix ];
            if( mrGrid . GetNumber( x, y ) == 0 ) {
                // Found an empty field
                uint64_t const  mask = mrGrid . GetAllowedMask( x, y );
                if( mask == 0 ) {
                    // We are stuck: this field is empty but there is no allowed number
                    return false;
                }
                // Try all allowed numbers
                for( int  num = 1; num <= mCount; ++num ) {
                    if( ( Allowed::NumberMask( num ) & mask ) && mrGrid . SetNumber( x, y, num ) ) {
                        // Do not use Solve() here because it does not restore fields on failure
                        if( FillFields( ix + 1, iy ) ) {
                            return true;
                        }
                        // Reset field for next test
                        mrGrid . SetNumber( x, y, 0 );
                    }
                }
                // No fitting number found
                return false;
            }
        }
        start_ix = 0;
    }

    // Unreachable: no empty field found but we are not complete?
    return false;
}

bool Generator::Instance::RemoveFields( int const  sI )
{
    // Check if already removed enough
    if( mrGrid . GetFilledFields() <= mGenerateOptions . GetFixedFields() ) {
        return true;
    }

    // Now try to find filled fields to clear
    std::auto_ptr< Grid >  p_backup( new Grid( mrGrid ) );
    int const  count = mCount * mCount;
    for( int  i = sI; i < count; ++i ) {
        int const  x = mFieldOrder[ i ] . first;
        int const  y = mFieldOrder[ i ] . second;
        if( mrGrid . GetNumber( x, y ) != 0 ) {
            // Found a filled field
            *p_backup = mrGrid;
            if( RemoveField( x, y ) ) {
                // Recursive try again
                if( RemoveFields( i + 1 ) ) {
                    return true;
                }
            }
            // Restore grid
            mrGrid = *p_backup;
        }
    }

    // Removed as much as possible
    return true;
}

bool Generator::Instance::RemoveField( int const  x, int const  y )
{
    Solver const  solver( mGenerateOptions );

    // Remove number
    mrGrid . SetNumber( x, y, 0 );
    bool  unique = solver . CheckUnique( mrGrid, x, y ) != 0;

    if( mGenerateOptions . IsSymmetric() ) {
        int const  c = mCount - 1;

        // Remove symmetrical field
        mrGrid . SetNumber( c - x, c - y, 0 );
        unique = unique && solver . CheckUnique( mrGrid, c - x, c - y );
    }

    // If unique solvable no more checks are needed
    if( unique ) return true;

    // Check for unique solution
    return solver . Analyse( mrGrid ) . GetSolution() == SolveResult::UniqueSolution;
}

void Generator::Instance::FixFields()
{
    for( int  y = 0; y < mCount; ++y ) {
        for( int  x = 0; x < mCount; ++x ) {
            mrGrid . SetFixed( x, y, mrGrid . GetNumber( x, y ) != 0 );
        }
    }
}

}
