/*
sudoku - Create and play sudoku

Copyright (C) 2007 Sven Heinzel <Sven.Heinzel@gmx.net>
 
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "sudoku/sudoku.h"
#include "sudoku/Solver.h"

namespace Sudoku {

bool Solver::SolveOne( Grid  &rGrid ) const
{
    Solver::Instance  instance( mSolveOptions, rGrid );
    int const  count = rGrid . GetCount();
    do {
        // Find first unique field
        for( int  y = 0; y < count; ++y ) {
            for( int  x = 0; x < count; ++x ) {
                // Empty field
                if( rGrid . GetNumber( x, y ) == 0 ) {
                    int const  num = instance . CheckUnique( x, y );
                    // Solve unique number in the original grid
                    if( num != 0 ) return rGrid . SetNumber( x, y, num );
                }
            }
        }
        // If no unique was found retry as long we can simplify allowed numbers
    } while( instance . SimplifyAllowed() );

    // Nothing found
    return false;
}

Grid Solver::FindTips( Grid const  &rGrid, bool const  simplifyAllowed ) const
{
    // Two copies to avoid finding too much uniques
    std::auto_ptr< Grid >  p_grid( new Grid( rGrid ) );
    std::auto_ptr< Grid >  p_tips( new Grid( rGrid ) );

    // Use one grid for checking and the other for modifying
    Solver::Instance  instance( mSolveOptions, *p_grid );
    if( simplifyAllowed ) {
        while( instance . SimplifyAllowed() );
    }

    // Modify only in the tips grid
    int const  count = p_grid -> GetCount();
    for( int  y = 0; y < count; ++y ) {
        for( int  x = 0; x < count; ++x ) {
            // Empty field
            if( p_grid -> GetNumber( x, y ) == 0 ) {
                int const  num = instance . CheckUnique( x, y );
                // Solve unique number in the tips grid
                if( num != 0 ) p_tips -> SetNumber( x, y, num );
            }
        }
    }

    // Return modified grid
    return *p_tips;
}

Allowed Solver::SimplifyAllowed( Grid const  &rGrid ) const
{
    // Solve a copy
    std::auto_ptr< Grid >  p_grid( new Grid( rGrid ) );
    Solver::Instance  instance( mSolveOptions, *p_grid );
    while( instance . SimplifyAllowed() );
    return instance . GetAllowed();
}

bool Solver::SolveAll( Grid  &rGrid ) const
{
    Solver::Instance  instance( mSolveOptions, rGrid );
    return instance . Solve( false );
}

SolveResult Solver::Analyse( Grid const  &rGrid ) const
{
    // Use a copy for analysing
    std::auto_ptr< Grid >  p_grid( new Grid( rGrid ) );
    Solver::Instance  instance( mSolveOptions, *p_grid );
    instance . Solve( true );
    return instance . GetSolveResult();
}

int Solver::CheckUnique( Grid const  &rGrid, int const  x, int const  y ) const
{
    // Use a copy for check
    std::auto_ptr< Grid >  p_grid( new Grid( rGrid ) );
    Solver::Instance  instance( mSolveOptions, *p_grid );
    return instance . CheckUnique( x, y );
}

}
