/*
sudoku - Create and play sudoku

Copyright (C) 2007 Sven Heinzel <Sven.Heinzel@gmx.net>
 
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "sudoku/sudoku.h"
#include "sudoku/Grid.h"
#include "sudoku/Solver.h"

namespace Sudoku {

bool Solver::Instance::Solve(
    bool const  checkMultiple
)
{
    // Continue solving while something can be solved
    while( ! mrGrid . IsComplete() && ( SolveUniques() || SimplifyAllowed() ) );

    // If solution is found it is unique
    if( mrGrid . IsComplete() ) {
        IncreaseSolution( SolveResult::UniqueSolution );
        return true;
    }

    // Not finished so use back track (if allowed)
    return mSolveOptions . IsBacktrackAllowed() && BackTrack( checkMultiple );
}

bool Solver::Instance::BackTrack(
    bool const  checkMultiple
)
{
    // Check if already completed
    if( mrGrid . IsComplete() ) {
        IncreaseSolution( SolveResult::UniqueSolution );
        return true;
    }

    // Check if unsolvable
    if( ! IsValid() ) {
        return false;
    }

    // Mark back tracking used
    mSolveResult . SetBacktrackAllowed( true );

    // Now try to find empty fields to fill
    for( int  y = 0; y < mCount; ++y ) {
        for( int  x = 0; x < mCount; ++x ) {
            if( mrGrid . GetNumber( x, y ) == 0 ) {
                // Found an empty field
                uint64_t const  mask = GetAllowedMask( x, y );
                if( mask == 0 ) {
                    // We are stuck: this field is empty but there is no allowed number
                    return false;
                }
                // Try all allowed numbers
                std::auto_ptr< Grid >  p_backup( new Grid( mrGrid ) );
                std::auto_ptr< Allowed >  p_backup_allowed( new Allowed( mAllowed ) );
                for( int  num = 1; num <= mCount; ++num ) {
                    if( ( Allowed::NumberMask( num ) & mask ) && SetNumber( x, y, num ) ) {
                        if( Solve( checkMultiple ) ) {
                            // Solved and not interested in more
                            if( ! checkMultiple || mSolveResult . GetSolution() >= SolveResult::MultipleSolutions ) {
                                return true;
                            }
                            // Check for multiple solution
                            if( ++mFoundSolutions > 1 ) {
                                IncreaseSolution( SolveResult::MultipleSolutions );
                                return true;
                            }
                        }
                        // Reset field for next test
                        mrGrid = *p_backup;
                        mAllowed = *p_backup_allowed;
                    }
                }
                // No fitting number found
                return false;
            }
        }
    }

    // Unreachable: no empty field found but we are not complete?
    return false;
}

bool Solver::Instance::SolveUniques()
{
    bool  found = false;

    for( int  y = 0; y < mCount; ++y ) {
        for( int  x = 0; x < mCount; ++x ) {
            // Empty field
            if( mrGrid . GetNumber( x, y ) == 0 ) {
                int const  num = CheckUnique( x, y );
                // Solve unique number
                if( num != 0 ) found = SetNumber( x, y, num ) || found;
            }
        }
    }

    return found;
}

int Solver::Instance::CheckUnique( int const  x, int const  y ) const
{
    // 1st try: only check unique per field
    uint64_t const  mask = GetAllowedMask( x, y );
    for( int  num = 1; num <= mCount; ++num ) {
        uint64_t const  num_mask = Allowed::NumberMask( num );
        if( num_mask == mask ) {
            // Unique per field
            return num;
        }
    }

    // 2nd try: check unique per row, column
    uint64_t const  unique_col = UniqueMaskColumn( x );
    uint64_t const  unique_row = UniqueMaskRow( y );
    uint64_t const  unique_box = UniqueMaskBox( mrGrid . BoxIndex( x, y ) );
    for( int  num = 1; num <= mCount; ++num ) {
        uint64_t const  num_mask = Allowed::NumberMask( num );
        bool const  allowed =    ( num_mask & mask ) // number is allowed...
                              && ( // and...
                                     ( num_mask & unique_col )  // unique per column
                                  || ( num_mask & unique_row )  // unique per row
                                  || ( num_mask & unique_box )  // unique per box
                                 );
        if( allowed ) return num;
    }

    // Nothing found
    return 0;
}

bool Solver::Instance::CheckPairCol( int const  x, int const  level )
{
    bool  found = false;

    // Only read allowed values once
    uint64_t  allowed_col[ 64 ];
    for( int  y = 0; y < mCount; ++y ) {
        allowed_col[ y ] = GetAllowedMask( x, y );
    }

    for( int  search = 0; search < mCount; ++search ) {
        uint64_t const  search_col = allowed_col[ search ];
        int const  count_col = Allowed::CountAllowed( search_col );

        // Find pairs
        int  found_col = 0;
        for( int  test = 0; test < mCount; ++test ) {
            uint64_t const  test_col = allowed_col[ test ];

            if( count_col == level && test_col && search_col == ( search_col | test_col ) ) ++found_col;
        }

        // Remove impossible fields
        for( int  i = 0; i < mCount; ++i ) {
            uint64_t const  test_col = allowed_col[ i ];

            if( found_col > level && search_col == ( search_col | test_col ) ) {
                // Found more matches than possible => unsolvable
                // So simply remove all matching fields to mark this field as unsolvable
                mAllowed . SetFieldMask( x, i, 0 );
            } else if( found_col == level && search_col != ( search_col | test_col ) ) {
                uint64_t const  new_allowed = test_col & ~search_col;
                if( test_col != new_allowed ) {
                    found = true;
                    mAllowed . SetFieldMask( x, i, new_allowed );
                }
            }
        }
    }

    return found;
}

bool Solver::Instance::CheckPairRow( int const  y, int const  level )
{
    bool  found = false;

    // Only read allowed values once
    uint64_t  allowed_row[ 64 ];
    for( int  x = 0; x < mCount; ++x ) {
        allowed_row[ x ] = GetAllowedMask( x, y );
    }

    for( int  search = 0; search < mCount; ++search ) {
        uint64_t const  search_row = allowed_row[ search ];
        int const  count_row = Allowed::CountAllowed( search_row );

        // Find pairs
        int  found_row = 0;
        for( int  test = 0; test < mCount; ++test ) {
            uint64_t const  test_row = allowed_row[ test ];

            if( count_row == level && test_row && search_row == ( search_row | test_row ) ) ++found_row;
        }

        // Remove impossible fields
        for( int  i = 0; i < mCount; ++i ) {
            uint64_t const  test_row = allowed_row[ i ];

            if( found_row > level && search_row == ( search_row | test_row ) ) {
                // Found more matches than possible => unsolvable
                // So simply remove all matching fields to mark this field as unsolvable
                mAllowed . SetFieldMask( i, y, 0 );
            } else if( found_row == level && search_row != ( search_row | test_row ) ) {
                uint64_t const  new_allowed = test_row & ~search_row;
                if( test_row != new_allowed ) {
                    found = true;
                    mAllowed . SetFieldMask( i, y, new_allowed );
                }
            }
        }
    }

    return found;
}

bool Solver::Instance::CheckPairBox( int const  b, int const  level )
{
    bool  found = false;

    // Only read allowed values once
    int  i_box = 0;
    uint64_t  allowed_box[ 64 ];
    int  x_box[ 64 ];
    int  y_box[ 64 ];
    for( int  y = 0; y < mCount; ++y ) {
        for( int  x = 0; x < mCount; ++x ) {
            if( mrGrid . BoxIndex( x, y ) == b ) {
                allowed_box[ i_box ] = GetAllowedMask( x, y );
                x_box[ i_box ] = x;
                y_box[ i_box ] = y;
                ++i_box;
            }
        }
    }

    for( int  search = 0; search < mCount; ++search ) {
        uint64_t const  search_box = allowed_box[ search ];
        int const  count_box = Allowed::CountAllowed( search_box );

        // Find pairs
        int  found_box = 0;
        for( int  test = 0; test < mCount; ++test ) {
            uint64_t const  test_box = allowed_box[ test ];

            if( count_box == level && test_box && search_box == ( search_box | test_box ) ) ++found_box;
        }

        // Remove impossible fields
        for( int  i = 0; i < mCount; ++i ) {
            uint64_t const  test_box = allowed_box[ i ];

            if( found_box > level && search_box == ( search_box | test_box ) ) {
                // Found more matches than possible => unsolvable
                // So simply remove all matching fields to mark this field as unsolvable
                mAllowed . SetFieldMask( x_box[ i ], y_box[ i ], 0 );
            } else if( found_box == level && search_box != ( search_box | test_box ) ) {
                uint64_t const  new_allowed = test_box & ~search_box;
                if( test_box != new_allowed ) {
                    found = true;
                    mAllowed . SetFieldMask( x_box[ i ], y_box[ i ], new_allowed );
                }
            }
        }
    }

    return found;
}

bool Solver::Instance::SimplifyAllowed()
{
    for( int  i = 2; i <= mSolveOptions . GetMaxPairSolving(); ++i ) {
        if( CheckPair( i ) ) return true;
    }

    // Nothing simplified
    return false;
}

bool Solver::Instance::CheckPair( int const  level )
{
    bool  found = false;

    for( int  place = 0; place < mCount; ++place ) {
#if 0
        found = CheckPairCol( place, level ) || found;
        found = CheckPairRow( place, level ) || found;
        found = CheckPairBox( place, level ) || found;
#else
        found = found || CheckPairCol( place, level );
        found = found || CheckPairRow( place, level );
        found = found || CheckPairBox( place, level );
#endif
    }

    if( found ) IncreaseMaxPairSolving( level );
    return found;
}

}
